#!/usr/bin/env python3
from typing import Any, Iterator
from urllib.parse import quote_plus

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroIam(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO IAM REST API.
    See https://core.arago.co/help/specs/?url=definitions/iam.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='iam',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations against the IAM API
    ###############################################################################################################

    ###############################################################################################################
    # Accounts
    ###############################################################################################################

    def create_account(self, data: dict, import_flag: bool = None) -> dict:
        """
        create an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/post_accounts

        :param import_flag: Default is false
        :param data: The dict with the account data.
        :return: Dict with the result
        """
        query = {
            "import": import_flag
        }

        url = self.endpoint + "/accounts" + self._get_query_part(query)
        return self.post(url, data)

    def update_account(self, account_id: str, data: dict) -> dict:
        """
        update an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/post_accounts__id_

        :param account_id: ogit/_id of the ogit/Auth/Account
        :param data: The dict with the account data.
        :return: Dict with the result
        """
        url = self.endpoint + "/accounts/" + quote_plus(account_id)
        return self.post(url, data)

    def get_account(self, account_id: str, profile: bool = None) -> dict:
        """
        gets an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/get_accounts__id_

        :param profile: return account with profile. Default false.
        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result
        """
        query = {
            "profile": profile
        }

        url = self.endpoint + "/accounts/" + quote_plus(account_id) + self._get_query_part(query)
        return self.get(url)

    def delete_account(self, account_id: str) -> dict:
        """
        deletes an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/delete_accounts__id_

        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result
        """
        url = self.endpoint + "/accounts/" + quote_plus(account_id)
        return self.delete(url)

    def activate_account(self, node_id: str) -> dict:
        """
        activates an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/patch_accounts__id__activate

        :param node_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result
        """
        url = self.endpoint + "/accounts/" + quote_plus(node_id) + "/activate"
        return self.patch(url, data=None)

    def get_account_avatar(self, account_id: str) -> Iterator[bytes]:
        """
        gets avatar of account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/get_accounts__id__avatar

        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Binary content of the avatar
        """
        url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/avatar"
        return self.get_binary(url)

    def put_account_avatar(self, account_id: str, data: Any, content_type: str = 'image/png') -> str:
        """
        sets the avatar of account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/put_accounts__id__avatar

        :param account_id: ogit/_id of the ogit/Auth/Account
        :param data: Binary data for image of avatar.
        :param content_type: Content-Type. Default: image/png
        :return: The result payload / size of the avatar in bytes.
        """
        url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/avatar"
        return self.put_binary(url, data, content_type=content_type)

    def deactivate_account(self, account_id: str, reason: str) -> dict:
        """
        deactivates an account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/patch_accounts__id__deactivate

        :param reason: reason of deactivation of an ogit/Auth/Account.
                       Available values : UserDeactivated, AdminDeactivated, PasswordExpired, AutoDeactivated, None
        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result
        """
        query = {
            "reason": reason
        }

        url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/deactivate" + self._get_query_part(query)
        return self.patch(url, data=None)

    def put_account_password(self, account_id: str, data: dict) -> dict:
        """
        set password of account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/put_accounts__id__password

        :param account_id: ogit/_id of the ogit/Auth/Account
        :param data: The dict with the new password data.
        :return: Dict with the result
        """
        url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/password"
        return self.post(url, data)

    def get_account_profile(self, account_id: str = None, profile_id: str = None) -> dict:
        """
        get profile of account

        You need to specify either profile_id or account_id.

          With account_id: https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/get_accounts__id__profile

          With profile_id: https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/get_accounts_profile__profileId_

        :param profile_id: ogit/_id of the ogit/Auth/AccountProfile
        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result or an empty dict if neither account_id nor profile_id are given.
        """
        if account_id:
            url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/profile"
        elif profile_id:
            url = self.endpoint + "/accounts/profile/" + quote_plus(profile_id)
        else:
            return {}

        return self.get(url)

    def get_account_teams(self, account_id: str, include_virtual: bool = None) -> dict:
        """
        get teams of account
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/get_accounts__id__teams

        :param include_virtual: return virtual teams in output. Default false.
        :param account_id: ogit/_id of the ogit/Auth/Account
        :return: Dict with the result
        """
        query = {
            "include_virtual": include_virtual
        }

        url = self.endpoint + "/accounts/" + quote_plus(account_id) + "/teams" + self._get_query_part(query)
        return self.get(url)

    def update_account_profile(self, profile_id: str, data: dict) -> dict:
        """
        updates an account profile
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Account/post_accounts_profile__profileId_

        :param data: Dict with the new profile data.
        :param profile_id: ogit/_id of the ogit/Auth/AccountProfile
        :return: Dict with the result
        """
        url = self.endpoint + "/accounts/profile/" + quote_plus(profile_id)
        return self.post(url, data)

    ###############################################################################################################
    # DataScope
    ###############################################################################################################

    def create_scope(self, data: dict) -> dict:
        """
        creates a DataScope
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_DataScope/post_scope

        :param data: Dict with the new scope data.
        :return: Dict with the result
        """
        url = self.endpoint + "/scope"
        return self.post(url, data)

    def get_scope(self, scope_id: str) -> dict:
        """
        retrieves a DataScope
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_DataScope/get_scope__id_

        :param scope_id: ogit/_id of the ogit/Auth/DataScope
        :return: Dict with the result
        """
        url = self.endpoint + "/scope/" + quote_plus(scope_id)
        return self.get(url)

    def update_scope(self, scope_id: str, data: dict) -> dict:
        """
        updates a DataScope
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_DataScope/put_scope__id_

        :param scope_id: ogit/_id of the ogit/Auth/DataScope
        :param data: Dict with the new scope data.
        :return: Dict with the result
        """
        url = self.endpoint + "/scope/" + quote_plus(scope_id)
        return self.put(url, data)

    ###############################################################################################################
    # Organization
    ###############################################################################################################

    def create_organization(self, data: dict) -> dict:
        """
        creates an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/post_organization

        :param data: Dict with the new organization data.
        :return: Dict with the result
        """
        url = self.endpoint + "/organization"
        return self.post(url, data)

    def get_organization_configuration(self, organization_id: str) -> dict:
        """
        gets configuration of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__configuration

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/configuration"
        return self.get(url)

    def update_organization_configuration(self, organization_id: str, data: dict) -> dict:
        """
        updates configuration of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/put_organization__id__configuration

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :param data: Dict with the new organization data.
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/configuration"
        return self.put(url, data)

    def get_organization_datasets(self, organization_id: str) -> dict:
        """
        gets datasets of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__datasets

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/datasets"
        return self.get(url)

    def get_organization_domains(self, organization_id: str) -> dict:
        """
        gets domains of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__domains

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/domains"
        return self.get(url)

    def get_organization_roleassignments(self, organization_id: str) -> dict:
        """
        gets roleassignments of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__roleassignments

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/roleassignments"
        return self.get(url)

    def get_organization_scopes(self, organization_id: str) -> dict:
        """
        gets scopes of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__scopes

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/scopes"
        return self.get(url)

    def get_organization_teams(self, organization_id: str) -> dict:
        """
        gets teams of an Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__teams

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Dict with the result
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/teams"
        return self.get(url)

    def get_organization_avatar(self, organization_id: str) -> Iterator[bytes]:
        """
        gets avatar of Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/get_organization__id__avatar

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :return: Binary content of the avatar
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/avatar"
        return self.get_binary(url)

    def put_organization_avatar(self, organization_id: str, data: Any, content_type: str = 'image/png') -> str:
        """
        sets the avatar of Auth Organization
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Organization/put_organization__id__avatar

        :param organization_id: ogit/_id of the ogit/Auth/Organization
        :param data: Binary data for image of avatar.
        :param content_type: Content-Type. Default: image/png
        :return: The result payload / size of the avatar in bytes.
        """
        url = self.endpoint + "/organization/" + quote_plus(organization_id) + "/avatar"
        return self.put_binary(url, data, content_type=content_type)

    ###############################################################################################################
    # Org Domain
    ###############################################################################################################

    def create_domain(self, data: dict) -> dict:
        """
        creates an Org Domain
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_OrgDomain/post_domain

        :param data: Dict with the new domain data.
        :return: Dict with the result
        """
        url = self.endpoint + "/domain"
        return self.post(url, data)

    def get_domain(self, domain_id: str) -> dict:
        """
        retrieves an Org Domain
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_OrgDomain/get_domain__id_

        :param domain_id: ogit/_id of the ogit/Auth/OrgDomain
        :return: Dict with the result
        """
        url = self.endpoint + "/domain/" + quote_plus(domain_id)
        return self.get(url)

    def delete_domain(self, domain_id: str) -> dict:
        """
        deletes an Org Domain
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_OrgDomain/delete_domain__id_

        :param domain_id: ogit/_id of the ogit/Auth/OrgDomain
        :return: Dict with the result
        """
        url = self.endpoint + "/domain/" + quote_plus(domain_id)
        return self.delete(url)

    def get_domain_organization(self, domain_id: str) -> dict:
        """
        retrieves an Organization of Org Domain
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_OrgDomain/get_domain__id__organization

        :param domain_id: ogit/_id of the ogit/Auth/OrgDomain
        :return: Dict with the result
        """
        url = self.endpoint + "/domain/" + quote_plus(domain_id) + "/organization"
        return self.get(url)

    ###############################################################################################################
    # Auth Role
    ###############################################################################################################

    def create_role(self, data: dict) -> dict:
        """
        creates an Auth Role
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Role/post_role

        :param data: Dict with the new role data.
        :return: Dict with the result
        """
        url = self.endpoint + "/role"
        return self.post(url, data)

    def update_role(self, role_id: str, data: dict) -> dict:
        """
        updates an Auth Role
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Role/put_role__id_

        :param role_id: ogit/_id of the ogit/Auth/Role
        :param data: Dict with the new role data.
        :return: Dict with the result
        """
        url = self.endpoint + "/role/" + quote_plus(role_id)
        return self.put(url, data)

    def get_role(self, role_id: str) -> dict:
        """
        retrieves an Auth Role
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Role/get_role__id_

        :param role_id: ogit/_id of the ogit/Auth/Role
        :return: Dict with the result
        """
        url = self.endpoint + "/role/" + quote_plus(role_id)
        return self.get(url)

    def delete_role(self, role_id: str) -> dict:
        """
        deletes an Auth Role
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Role/delete_role__id_

        :param role_id: ogit/_id of the ogit/Auth/Role
        :return: Dict with the result
        """
        url = self.endpoint + "/role/" + quote_plus(role_id)
        return self.delete(url)

    def get_roles(self) -> dict:
        """
        retrieves public roles
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Role/get_roles

        :return: Dict with the result
        """
        url = self.endpoint + "/roles"
        return self.get(url)

    ###############################################################################################################
    # Auth RoleAssignment
    ###############################################################################################################

    def create_roleassignment(self, data: dict) -> dict:
        """
        creates an Auth RoleAssignment
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_RoleAssignment/post_roleassignment

        :param data: Dict with the new roleassignment data.
        :return: Dict with the result
        """
        url = self.endpoint + "/roleassignment"
        return self.post(url, data)

    def get_roleassignment(self, roleassignment_id: str) -> dict:
        """
        retrieves an Auth RoleAssignment
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_RoleAssignment/get_roleassignment__id_

        :param roleassignment_id: ogit/_id of the ogit/Auth/RoleAssignment
        :return: Dict with the result
        """
        url = self.endpoint + "/roleassignment/" + quote_plus(roleassignment_id)
        return self.get(url)

    def delete_roleassignment(self, roleassignment_id: str) -> dict:
        """
        deletes an Auth RoleAssignment
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_RoleAssignment/delete_roleassignment__id_

        :param roleassignment_id: ogit/_id of the ogit/Auth/RoleAssignment
        :return: Dict with the result
        """
        url = self.endpoint + "/roleassignment/" + quote_plus(roleassignment_id)
        return self.delete(url)

    ###############################################################################################################
    # Auth Team
    ###############################################################################################################

    def create_team(self, data: dict) -> dict:
        """
        creates an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/post_team

        :param data: Dict with the new team data.
        :return: Dict with the result
        """
        url = self.endpoint + "/team"
        return self.post(url, data)

    def update_team(self, team_id: str, data: dict) -> dict:
        """
        updates an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/put_team__id_

        :param team_id: ogit/_id of the ogit/Auth/Team
        :param data: Dict with the new team data.
        :return: Dict with the result
        """
        url = self.endpoint + "/team/" + quote_plus(team_id)
        return self.put(url, data)

    def get_team(self, team_id: str) -> dict:
        """
        retrieves an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/get_team__id_

        :param team_id: ogit/_id of the ogit/Auth/Team
        :return: Dict with the result
        """
        url = self.endpoint + "/team/" + quote_plus(team_id)
        return self.get(url)

    def delete_team(self, team_id: str) -> dict:
        """
        deletes an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/delete_team__id_

        :param team_id: ogit/_id of the ogit/Auth/Team
        :return: Dict with the result
        """
        url = self.endpoint + "/team/" + quote_plus(team_id)
        return self.delete(url)

    def get_team_members(self, team_id: str, profile: bool = None) -> dict:
        """
        gets members of an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/get_team__id__members

        :param profile: return account with profile. Default False.
        :param team_id: ogit/_id of the ogit/Auth/Team
        :return: Dict with the result
        """

        query = {
            "profile": profile
        }

        url = self.endpoint + "/team/" + quote_plus(team_id) + "/members" + self._get_query_part(query)
        return self.get(url)

    def add_team_members(self, team_id: str, data: dict) -> dict:
        """
        adds members to an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/post_team__id__members_add

        :param team_id: ogit/_id of the ogit/Auth/Team
        :param data: dict with the account data
        :return: Dict with the result
        """

        url = self.endpoint + "/team/" + quote_plus(team_id) + "/members/add"
        return self.post(url, data)

    def remove_team_members(self, team_id: str, data: dict) -> dict:
        """
        adds members to an Auth Team
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_Auth_Team/post_team__id__members_remove

        :param team_id: ogit/_id of the ogit/Auth/Team
        :param data: dict with the account data
        :return: Dict with the result
        """

        url = self.endpoint + "/team/" + quote_plus(team_id) + "/members/remove"
        return self.post(url, data)

    ###############################################################################################################
    # DataSet
    ###############################################################################################################

    def create_dataset(self, data: dict) -> dict:
        """
        creates a DataSet
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_DataSet/post_dataset

        :param data: Dict with the new dataset data.
        :return: Dict with the result
        """
        url = self.endpoint + "/dataset"
        return self.post(url, data)

    def update_dataset(self, dataset_id: str, data: dict) -> dict:
        """
        updates a DataSet
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_DataSet/put_dataset__id_

        :param dataset_id: ogit/_id of the ogit/Auth/DataSet
        :param data: Dict with the new dataset data.
        :return: Dict with the result
        """
        url = self.endpoint + "/dataset/" + quote_plus(dataset_id)
        return self.put(url, data)

    def get_dataset(self, dataset_id: str) -> dict:
        """
        retrieves a DataSet
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_DataSet/get_dataset__id_

        :param dataset_id: ogit/_id of the ogit/Auth/DataSet
        :return: Dict with the result
        """
        url = self.endpoint + "/dataset/" + quote_plus(dataset_id)
        return self.get(url)

    def delete_dataset(self, dataset_id: str) -> dict:
        """
        deletes a DataSet
        https://core.arago.co/help/specs/?url=definitions/iam.yaml#/[Management]_DataSet/delete_dataset__id_

        :param dataset_id: ogit/_id of the ogit/Auth/DataSet
        :return: Dict with the result
        """
        url = self.endpoint + "/dataset/" + quote_plus(dataset_id)
        return self.delete(url)



import json
import logging
import threading
import time
from abc import abstractmethod
from datetime import datetime
from enum import Enum
from typing import Optional, Union, Dict

from apscheduler.job import Job
from apscheduler.jobstores.base import JobLookupError
from apscheduler.schedulers.background import BackgroundScheduler
from websocket import WebSocketApp

from hiro_graph_client.clientlib import AbstractTokenApiHandler
from hiro_graph_client.websocketlib import AbstractAuthenticatedWebSocketHandler

logger = logging.getLogger(__name__)
""" The logger for this module """


###############################################################################################################
# Message models
###############################################################################################################

class ActionMessageType(str, Enum):
    ACKNOWLEDGED = 'acknowledged'
    NEGATIVE_ACKNOWLEDGED = 'negativeAcknowledged'
    SEND_ACTION_RESULT = 'sendActionResult'
    SUBMIT_ACTION = 'submitAction'
    ERROR = 'error'
    CONFIG_CHANGED = 'configChanged'


class AbstractActionHandlerMessage:
    """
    The structure of an incoming action message
    """
    type: ActionMessageType
    """ static type name of field 'type' in the json message """

    def __init__(self, **kwargs):
        pass

    def __str__(self) -> str:
        """ Create a JSON string representation of the message """
        result = {"type": self.type}
        result.update(vars(self))
        return json.dumps(result)

    @classmethod
    def parse(cls, dict_message: dict):
        pass


class AbstractActionHandlerIdMessage(AbstractActionHandlerMessage):
    """
    The structure of an incoming action message with id
    """
    id: str

    def __init__(self,
                 action_id: str,
                 **kwargs):
        """
        Constructor

        :param action_id: ID. Might be None for messages without an ID.
        """
        super().__init__(**kwargs)
        if not action_id:
            raise ValueError("{}: parameter 'action_id' required.".format(self.__class__.__name__))
        self.id = action_id


class AbstractActionHandlerCodeMessage(AbstractActionHandlerMessage):
    """
    The structure of an incoming action message with code and message
    """
    code: int
    message: str

    def __init__(self,
                 code: int,
                 message: str,
                 **kwargs):
        """
        Constructor

        :param code: Code number
        :param message: Message string
        """
        super().__init__(**kwargs)
        self.code = code
        self.message = message


class ActionHandlerSubmit(AbstractActionHandlerIdMessage):
    handler: str
    capability: str
    parameters: dict
    timeout: int
    """Milliseconds"""

    _expires_at: int
    """Milliseconds"""

    type = ActionMessageType.SUBMIT_ACTION

    def __init__(self,
                 action_id: str,
                 handler: str,
                 capability: str,
                 parameters: dict,
                 timeout: int):
        """
        Constructor

        :param action_id: ID
        :param handler: Handler name
        :param capability: Capability name
        :param parameters: Parameter dict
        :param timeout: Timeout in milliseconds
        """
        super().__init__(action_id=action_id)
        self.handler = handler
        self.capability = capability
        self.parameters = parameters
        self.timeout = timeout

        self._expires_at = int(time.time_ns() / 1000000 + self.timeout)

    @classmethod
    def parse(cls, dict_message: dict):
        return cls(
            action_id=dict_message.get('id'),
            handler=dict_message.get('handler'),
            capability=dict_message.get('capability'),
            parameters=dict_message.get('parameters'),
            timeout=dict_message.get('timeout')
        )

    @property
    def expires_at(self):
        return self._expires_at


class ActionHandlerResult(AbstractActionHandlerIdMessage):
    result: dict

    type = ActionMessageType.SEND_ACTION_RESULT

    def __init__(self,
                 action_id: str,
                 result: Union[dict, str]):
        """
        Constructor

        :param action_id: ID
        :param result: Result data
        """
        super().__init__(action_id=action_id)
        self.result = result

    @classmethod
    def parse(cls, dict_message: dict):
        return cls(
            action_id=dict_message.get('id'),
            result=dict_message.get('result')
        )

    def stringify_result(self) -> str:
        return json.dumps({
            "type": self.type,
            "id": self.id,
            "result": json.dumps(self.result)
        })


class ActionHandlerAck(AbstractActionHandlerIdMessage, AbstractActionHandlerCodeMessage):
    type = ActionMessageType.ACKNOWLEDGED

    def __init__(self,
                 action_id: str,
                 code: int,
                 message: str):
        """
        Constructor

        :param action_id: ID
        :param code: Ack code
        :param message: Ack message
        """
        super().__init__(action_id=action_id, code=code, message=message)

    @classmethod
    def parse(cls, dict_message: dict):
        return cls(
            action_id=dict_message.get('id'),
            code=dict_message.get('code'),
            message=dict_message.get('message')
        )


class ActionHandlerNack(AbstractActionHandlerIdMessage, AbstractActionHandlerCodeMessage):
    type = ActionMessageType.NEGATIVE_ACKNOWLEDGED

    def __init__(self,
                 action_id: str,
                 code: int,
                 message: str):
        """
        Constructor

        :param action_id: ID
        :param code: Nack code
        :param message: Nack message
        """
        super().__init__(action_id=action_id, code=code, message=message)

    @classmethod
    def parse(cls, dict_message: dict):
        return cls(
            action_id=dict_message.get('id'),
            code=dict_message.get('code'),
            message=dict_message.get('message')
        )


class ActionHandlerError(AbstractActionHandlerCodeMessage):
    type = ActionMessageType.ERROR

    def __init__(self, code: int, message: str):
        """
        Constructor

        :param code: Error code
        :param message: Error message
        """
        super().__init__(code=code, message=message)

    @classmethod
    def parse(cls, dict_message: dict):
        return cls(
            code=dict_message.get('code'),
            message=dict_message.get('message')
        )


class ActionHandlerConfigChanged(AbstractActionHandlerMessage):
    type = ActionMessageType.CONFIG_CHANGED

    def __init__(self):
        super().__init__()

    @classmethod
    def parse(cls, dict_message: dict):
        return cls()


class ActionHandlerMessageParser:
    """
    Create an ActionHandlerMessage from a string message or dict.
    """

    message: dict

    def __init__(self, message: Union[dict, str]):
        """
        Create an ActionHandlerMessage from a string message or dict. Return None if no message can be parsed.

        :param message: Message as str or dict
        """
        self.message = json.loads(message) if isinstance(message, str) else message

    def parse(self) -> Optional[AbstractActionHandlerMessage]:
        """
        Create the ActionHandlerMessage

        :return: Instance of a child of AbstractActionHandlerMessage or None if *self.message* is empty.
        :raise UnknownActionException: If the type of *self.message* is unknown.
        """
        if self.message:
            message_type = self.message.get('type')

            try:
                action_type = ActionMessageType(message_type)
            except ValueError as err:
                raise UnknownActionException(message_type=message_type, error_id=self.message.get('id')) from err

            if action_type == ActionMessageType.SUBMIT_ACTION:
                return ActionHandlerSubmit.parse(self.message)
            if action_type == ActionMessageType.SEND_ACTION_RESULT:
                return ActionHandlerResult.parse(self.message)
            if action_type == ActionMessageType.ACKNOWLEDGED:
                return ActionHandlerAck.parse(self.message)
            if action_type == ActionMessageType.NEGATIVE_ACKNOWLEDGED:
                return ActionHandlerNack.parse(self.message)
            if action_type == ActionMessageType.CONFIG_CHANGED:
                return ActionHandlerConfigChanged.parse(self.message)
            if action_type == ActionMessageType.ERROR:
                return ActionHandlerError.parse(self.message)

        return None


###############################################################################################################
# Special ActionHandlerMessage storage
###############################################################################################################

class ActionItem:
    """
    An item for the *ExpiringStore* which carries its timeout and retries
    """

    message: AbstractActionHandlerIdMessage
    expires_at: int
    """ Epoch time after which this item is invalid """
    retries: int
    """ Amount of retries this item can be get via the ActionStore.retryGet(). Default is 4 """

    _scheduled_job: Job

    def __init__(self, job: Job, message: AbstractActionHandlerIdMessage, expires_at: int, retries: int = 4):
        """
        Constructor

        :param job: Job of the scheduler
        :param message: AbstractActionHandlerMessage
        :param expires_at: Expiry of the message in epoch in ms
        :param retries: Retry counter for *ActionStore.retry_get*. Defaults to 4.
        """
        self._scheduled_job = job
        self.message = message
        self.expires_at = expires_at
        self.retries = retries

    def __del__(self):
        self.remove()

    def remove(self):
        """
        Remove the job from the scheduler on deletion.
        """
        try:
            self._scheduled_job.remove()
        except JobLookupError:
            pass


class ActionStore:
    """
    A thread-safe storage implementation with expiring entries
    """

    _store: Dict[str, ActionItem]
    _store_lock: threading.RLock

    _expiry_scheduler: BackgroundScheduler

    def __init__(self):
        """
        Constructor. Starts the scheduler.
        """
        self._store = {}
        self._store_lock = threading.RLock()
        self._expiry_scheduler = BackgroundScheduler()

        if logging.root.level == logging.INFO:
            logging.getLogger('apscheduler').setLevel(logging.WARNING)

        # self.__expiry_scheduler.start()

    def __del__(self):
        """ Destructor. Shuts down the scheduler. """
        self.clear()
        self.stop_scheduler()

    def start_scheduler(self):
        if not self._expiry_scheduler.running:
            self._expiry_scheduler.start()

    def stop_scheduler(self):
        if self._expiry_scheduler.running:
            self._expiry_scheduler.shutdown()

    def __expiry_remove(self, message_id: str) -> None:
        """
        Remove a message from the store and log it on *logger.isEnabledFor(logging.INFO)*.

        :param message_id: Id of the message
        """
        with self._store_lock:
            if message_id in self._store:
                item = self._store.pop(message_id)
                logger.info("Discard %s %s because it expired.", item.message.type, item.message.id)
                item.remove()

    def add(self,
            expires_at: int,
            message: AbstractActionHandlerIdMessage,
            retries: int = 4) -> None:
        """
        Add message to the store and start its expiry timer.

        :param expires_at: Epoch in ms when the message expires.
        :param message: The message itself.
        :param retries: Retries allowed when *self.retry_get* is used.
        :raise ActionItemExpired: When the expires_at has expired already.
        :raise ActionItemExists: When the id of the message already exists in the store
        """

        if expires_at - (time.time_ns() / 1000000) < 0:
            raise ActionItemExpired(message.id, message.type)

        with self._store_lock:
            if message.id in self._store:
                raise ActionItemExists(message.id, message.type)

            job: Job = self._expiry_scheduler.add_job(
                func=lambda message_id: self.__expiry_remove(message_id),
                kwargs={"message_id": message.id},
                trigger='date',
                run_date=datetime.fromtimestamp(expires_at / 1000))

            self._store[message.id] = ActionItem(job, message, expires_at, retries)

    def remove(self, message_id: str) -> None:
        """
        Remove a message from the store.

        :param message_id: Id of the message
        """
        with self._store_lock:
            if message_id in self._store:
                self._store.pop(message_id).remove()

    def get(self, message_id: str) -> Optional[AbstractActionHandlerIdMessage]:
        """
        Get a message from the store.

        :param message_id: Id of the message
        :return: The message or None if message_id is not present in the store.
        """
        with self._store_lock:
            item: ActionItem = self._store.get(message_id)
            return item.message if item else None

    def retry_get(self, message_id: str) -> Optional[AbstractActionHandlerIdMessage]:
        """
        Get a message from the store. Each time it is returned, its retries counter is decreased. If it reaches 0,
        the message gets deleted and None gets returned.

        :param message_id: Id of the message
        :return: The message or None if message_id is not present in the store or its retries counter is 0.
        """
        with self._store_lock:
            item: ActionItem = self._store.get(message_id)
            if not item:
                return None

            if item.retries:
                item.retries -= 1
                return item.message
            else:
                logger.info("Discard message %s %s because no retries left.",
                            item.message.type,
                            item.message.id)
                self.remove(message_id)
                return None

    def clear(self) -> None:
        """
        Clear the store. Remove all items from it.
        """
        with self._store_lock:
            for item in self._store.values():
                item.remove()
            self._store.clear()


###############################################################################################################
# Main handler class
###############################################################################################################

class AbstractActionWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
    """
    A handler for actions
    """

    submitStore: ActionStore = None
    resultStore: ActionStore = None

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 query_params: Dict[str, str] = None):
        """
        Constructor

        :param api_handler: The TokenApiHandler for this WebSocket.
        :param query_params: URL Query parameters for this specific websocket. Use Dict[str,str] only here - no boolean
                             Values.
        """
        super().__init__(api_handler,
                         'action-ws',
                         query_params=query_params)

        self.submitStore = ActionStore()
        self.resultStore = ActionStore()

    def __del__(self):
        if self.submitStore:
            self.submitStore.stop_scheduler()
        if self.resultStore:
            self.resultStore.stop_scheduler()

    def __finish_submit(self, action_id: str, action_handler_result: Optional[ActionHandlerResult]):
        """
        Sends the action_handler_result if it is not None and makes sure, that the submitAction is removed from the
        self.submitStore.

        :param action_id: ID of the action
        :param action_handler_result: Result for the submitAction. May be None.
        """
        try:
            if action_handler_result:
                logger.info('Sending %s (id: %s)', action_handler_result.type, action_handler_result.id)
                self.send(action_handler_result.stringify_result())
        finally:
            self.submitStore.remove(action_id)

    def start(self) -> None:
        super().start()
        self.submitStore.start_scheduler()
        self.resultStore.start_scheduler()

    def stop(self) -> None:
        self.submitStore.stop_scheduler()
        self.resultStore.stop_scheduler()
        super().stop()

    ###############################################################################################################
    # Websocket Events
    ###############################################################################################################

    def on_open(self, ws: WebSocketApp):
        pass

    def on_close(self, ws: WebSocketApp, code: int = None, reason: str = None):
        pass

    def on_error(self, ws: WebSocketApp, error: Exception):
        pass

    def on_message(self, ws: WebSocketApp, message: str):
        """
        Handle incoming action messages

        :param ws: WebSocketApp
        :param message: The message payload as str
        """
        action_message = None
        try:
            action_message = ActionHandlerMessageParser(message).parse()
            if not action_message:
                return

            if isinstance(action_message, ActionHandlerSubmit):
                logger.info('Handling "%s" (id: %s)', action_message.type, action_message.id)

                self.send(str(ActionHandlerAck(action_message.id, 200, "submitAction acknowledged")))

                try:
                    self.submitStore.add(action_message.expires_at, action_message)
                except (ActionItemExists, ActionItemExpired) as err:
                    logger.info(str(err))
                    return

                action_handler_result = self.resultStore.get(action_message.id)
                if isinstance(action_handler_result, ActionHandlerResult):
                    logger.info('Handling "%s" (id: %s): Already processed', action_message.type, action_message.id)
                    self.__finish_submit(action_message.id, action_handler_result)
                    return

                try:
                    self.on_submit_action(action_message.id, action_message.capability, action_message.parameters)
                except Exception as err:
                    logger.error(str(err))
                    self.send_action_result(action_id=action_message.id, result=None, code=500, message=str(err))

            elif isinstance(action_message, ActionHandlerResult):
                logger.info('Handling "%s" (id: %s)', action_message.type, action_message.id)

                self.send(str(ActionHandlerNack(action_message.id, 400, "sendActionResult rejected")))

            elif isinstance(action_message, ActionHandlerAck):
                logger.info('Handling "%s" (id: %s)', action_message.type, action_message.id)

                self.resultStore.remove(action_message.id)

            elif isinstance(action_message, ActionHandlerNack):
                logger.info('Handling "%s" (id: %s)', action_message.type, action_message.id)

                action_handler_result = self.resultStore.retry_get(action_message.id)
                if isinstance(action_handler_result, ActionHandlerResult):
                    time.sleep(1)
                    self.send(action_handler_result.stringify_result())

            elif isinstance(action_message, ActionHandlerConfigChanged):
                logger.info('Handling "%s"', action_message.type)
                self.on_config_changed()

            elif isinstance(action_message, ActionHandlerError):
                logger.warning('Received error message (code %s): %s',
                               action_message.code,
                               action_message.message)

            else:
                logger.error("Received unknown message: %s", message)

        except UnknownActionException as err:
            logger.error(str(err))
            if err.id:
                self.send(str(ActionHandlerNack(err.id, 400, str(err))))
        except Exception as err:
            logger.error(str(err))
            if isinstance(action_message, AbstractActionHandlerIdMessage):
                self.send(str(ActionHandlerNack(action_message.id, 500, str(err))))

    ###############################################################################################################
    # API methods
    ###############################################################################################################

    @abstractmethod
    def on_submit_action(self, action_id: str, capability: str, parameters: dict):
        """
        Handle incoming submitAction

        :param action_id: Id of the action
        :param capability: Capability of the submitAction
        :param parameters: Additional parameters for the capability
        """
        pass

    def send_action_result(self,
                           action_id: str,
                           result: Union[dict, str] = None,
                           code: int = None,
                           message: str = None,
                           result_params: dict = None) -> None:
        """
        Send the result of a submitAction. Thread-Safe!

        Will create a standard result dict with the parameters *result*, *code* or *message*. See parameter description
        for more.

        ::

            {
                "type": "sendActionResult",
                "id": action_id,
                "result": json.dumps({
                    "message": message,
                    "code": code,
                    "data": result
                })

            }

        If *result_params* is given, this parameter will be used directly while ignoring the other parameters.

        ::

            {
                "type": "sendActionResult",
                "id": action_id,
                "result": json.dumps(result_params)
            }

        :param action_id: Required. Id of the submitAction
        :param result: Result data. Either a dict (which will be serialized into a str), a str or None.
        :param code: Optional result code. If unset, a default will be used (200 if result is not empty, 204 otherwise).
        :param message: Optional result message. If unset, a default message will be used.
        :param result_params: Optional custom data to return as result value in message *sendActionResult*.
                              Overrides all other optional message parameters. Will be serialized to a string.
        """
        result_message = ActionHandlerResult(action_id, result)
        submit_message = self.submitStore.get(action_id)

        if not isinstance(submit_message, ActionHandlerSubmit):
            logger.info('Handling "%s" (id: %s): Submit not stored - maybe it has expired?',
                        result_message.type,
                        result_message.id)
            return

        if not result_params:
            if not result:
                result_params = {
                    "message": message if message else "Action successful (no data)",
                    "code": code if code else 204
                }
            else:
                result_params = {
                    "message": message if message else "Action successful",
                    "code": code if code else 200,
                    "data": json.dumps(result) if isinstance(result, dict) else result
                }

        action_handler_result = ActionHandlerResult(result_message.id, result_params)
        try:
            self.resultStore.add(submit_message.expires_at, action_handler_result)
        except ActionItemExists as err:
            logger.info(str(err))
            action_handler_result = self.resultStore.get(result_message.id)
        except ActionItemExpired as err:
            logger.info(str(err))
            action_handler_result = None

        self.__finish_submit(action_id, action_handler_result)

    @abstractmethod
    def on_config_changed(self):
        pass


###################################################################################################################
# Exceptions
###################################################################################################################

class ActionException(Exception):
    """
    When an unknown action is received
    """
    id: str
    type: str

    def __init__(self, error_id, message_type):
        self.id = error_id
        self.type = message_type


class UnknownActionException(ActionException):
    """
    When an unknown action is received
    """

    def __init__(self, error_id, message_type):
        super().__init__(error_id, message_type)

    def __str__(self):
        return 'Unknown ActionHandlerMessage "{}" (id: {})'.format(self.type, self.id)


class ActionItemExpired(ActionException):
    """
    When an action item for the ActionStore has expired
    """

    def __init__(self, error_id, message_type):
        super().__init__(error_id, message_type)

    def __str__(self):
        return 'Message has expired: "{}" (id: {})'.format(self.type, self.id)


class ActionItemExists(ActionException):
    """
    When an action item for the ActionStore already exists
    """

    def __init__(self, error_id, message_type):
        super().__init__(error_id, message_type)

    def __str__(self):
        return 'Message id already stored: "{}" (id: {})'.format(self.type, self.id)



#!/usr/bin/env python3
from typing import List, Dict, Union

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroVariables(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO Variables REST API.
    See https://core.arago.co/help/specs/?url=definitions/variables.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='variables',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations against the variable API
    ###############################################################################################################

    def create_variable(self,
                        data: dict,
                        subtype: str = None) -> dict:
        """
        Creates new variable

        HIRO REST query API: `PUT self._endpoint`

        :param data: Variable content. See
               https://core.arago.co/help/specs/?url=definitions/variables.yaml#/[Variables]/put_
        :param subtype: Query variable. Value of ogit/subType. Optional.
        :return: The result payload
        """
        query = {
            "subtype": subtype
        }

        url = self.endpoint + self._get_query_part(query)
        return self.put(url, data)

    def get_variable(self,
                     name: str) -> dict:
        """
        Get variable by name. (synonym for self.define(name))

        HIRO REST query API: `GET self._endpoint + '/define'`

        :param name Variable name. Required.
        :return: The result payload
        """
        return self.define(name)

    def define(self,
               name: str) -> dict:
        """
        Get variable by name.

        HIRO REST query API: `GET self._endpoint + '/define'`

        :param name Variable name. Required.
        :return: The result payload
        """
        query = {
            "name": name
        }

        url = self.endpoint + "/define" + self._get_query_part(query)
        return self.get(url)

    def like(self,
             name: str,
             description: str = None,
             subtype: str = None,
             full: bool = None) -> Union[List, Dict]:
        """
        Search for similar variables.

        HIRO REST query API: `GET self._endpoint + '/like'`

        :param name Variable name. Required.
        :param description: Search by variable description. Optional.
        :param subtype: Value of ogit/subType. Optional.
        :param full: Return full variable, not just a name. Optional. Default false.
        :return: The result payload. Either a list of dict or a dict with an error message.
        """
        query = {
            "name": name,
            "description": description,
            "subtype": subtype,
            "full": full
        }

        url = self.endpoint + "/like" + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        res_list: list = res['items']
        return res_list

    def suggest(self,
                name: str,
                subtype: str = None,
                full: bool = None) -> Union[List, Dict]:
        """
        Search for similar variables.

        HIRO REST query API: `GET self._endpoint + '/suggest'`

        :param name Variable name. Required.
        :param subtype: Value of ogit/subType. Optional.
        :param full: Return full variable, not just a name. Optional. Default false.
        :return: The result payload. Either a list of dict or a dict with an error message.
        """
        query = {
            "name": name,
            "subtype": subtype,
            "full": full
        }

        url = self.endpoint + "/suggest" + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        res_list: list = res['items']
        return res_list



#!/usr/bin/env python3

from typing import Iterator
from urllib.parse import quote_plus

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroApp(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO App REST API.
    See https://core.arago.co/help/specs/?url=definitions/app.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='app',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations
    ###############################################################################################################

    def get_app(self, node_id) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}'`

        :param node_id: ogit/_id of the node/vertex or edge.
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id)
        return self.get(url)

    def get_config(self) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/config'`. The token (internal or external) defines the config
        returned.

        :return: The result payload
        """
        url = self.endpoint + '/config'
        return self.get(url)

    def get_content(self, node_id, path) -> Iterator[bytes]:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}/content/{path}'`. Get the content of an application.

        :param node_id: ogit/_id of the node/vertex or edge.
        :param path: filename / path of the desired content.
        :return: The result payload generator over binary data.
        """
        url = self.endpoint + '/' + quote_plus(node_id) + '/content/' + quote_plus(path)
        yield from self.get_binary(url)

    def get_manifest(self, node_id) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}/manifest'`. Get the manifest of an application.

        :param node_id: ogit/_id of the node/vertex or edge.
        :return: The result payload - usually with a binary content.
        """
        url = self.endpoint + '/' + quote_plus(node_id) + '/manifest'
        return self.get(url)

    def get_desktop(self) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/desktop'`. List desktop applications.

        :return: The result payload - usually with a binary content.
        """
        url = self.endpoint + '/desktop'
        return self.get(url)



#!/usr/bin/env python3
import base64
import json
import logging
import os
import threading
import time
import urllib
from abc import abstractmethod
from typing import Optional, Any, Iterator, Union, Tuple
from urllib.parse import quote, urlencode

import backoff
import requests
import requests.adapters
import requests.packages.urllib3.exceptions

from hiro_graph_client.version import __version__

logger = logging.getLogger(__name__)
""" The logger for this module """

BACKOFF_ARGS = [
    backoff.expo,
    requests.exceptions.RequestException
]
BACKOFF_KWARGS = {
    'jitter': backoff.random_jitter,
    'giveup': lambda e: e.response is not None and e.response.status_code < 500
}


###################################################################################################################
# SSL Configuration
###################################################################################################################

class SSLConfig:
    verify: bool
    cert_file: str
    key_file: str
    ca_bundle_file: str

    """
    This class contains the configuration for SSL connections, like the files to use.
    """

    def __init__(self,
                 verify: bool = True,
                 cert_file: str = None,
                 key_file: str = None,
                 ca_bundle_file: str = None):
        """
        Configuration for SSL connections.

        :param verify: Verify connections at all. If just set to True without other parameters, defaults will be used.
        :param cert_file: (optional) Client certificate file (.pem).
        :param key_file: (optional) Key for the certificate file.
        :param ca_bundle_file: (optional) The ca_bundle for server certificate verification.
        """
        self.verify = verify
        self.cert_file = cert_file
        self.key_file = key_file
        self.ca_bundle_file = ca_bundle_file

    def get_verify(self):
        """
        Get verify parameter as expected by requests library.

        :return: True, False or a path to a ca_bundle.
        """
        if self.verify and self.ca_bundle_file:
            return self.ca_bundle_file
        return self.verify

    def get_cert(self):
        """
        Get cert parameter as expected by requests library.

        :return: Tuple of cert_file, key_file or just cert_file - which can be None.
        """
        if self.cert_file and self.key_file:
            return self.cert_file, self.key_file
        return self.cert_file


###################################################################################################################
# Root classes for API
###################################################################################################################

class AbstractAPI:
    """
    This abstract root class contains the methods for HTTP requests used by all API classes. Also contains several
    tool methods for handling headers, url query parts and response error checking.
    """

    _root_url: str = None

    _session: requests.Session = None

    accept_all_certs: bool = False

    ssl_config: SSLConfig

    _client_name: str = "hiro-graph-client"

    _max_tries: int = 2

    _timeout: int = 600

    def __init__(self,
                 root_url: str = None,
                 session: requests.Session = None,
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = None,
                 client_name: str = None,
                 ssl_config: SSLConfig = None,
                 log_communication_on_error: bool = None,
                 max_tries: int = None,
                 abstract_api=None):

        """
        Constructor

        :param root_url: Root uri of the HIRO API, like *https://core.arago.co*.
        :param session: The requests.Session object for the connection pool. Required.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
               is given with another value for "User-Agent". Default is "hiro-graph-client".
        :param ssl_config: Optional configuration for SSL connections. If this is omitted, the defaults of `requests`
               lib will be used.
        :param log_communication_on_error: Log socket communication when an error (status_code of HTTP Response) is
               detected. Default is not to do this.
        :param max_tries: Max tries for BACKOFF. Default is 2.
        :param abstract_api: Set all parameters by copying them from the instance given by this parameter. Overrides
               all other parameters.
        """
        self._root_url = getattr(abstract_api, '_root_url', root_url)
        self._session = getattr(abstract_api, '_session', session)

        if not self._root_url:
            raise ValueError("'root_url' must not be empty.")

        if not self._session:
            raise ValueError("'session' must not be empty.")

        self._proxies = getattr(abstract_api, '_proxies', proxies)
        self._raise_exceptions = getattr(abstract_api, '_raise_exceptions', raise_exceptions)
        self._timeout = getattr(abstract_api, '_timeout', timeout or self._timeout)
        self._log_communication_on_error = getattr(abstract_api, '_log_communication_on_error',
                                                   log_communication_on_error or False)

        self.ssl_config = getattr(abstract_api, 'ssl_config', ssl_config or SSLConfig())

        if not self.ssl_config.verify:
            AbstractAPI.accept_all_certs = True
            requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)

        self._client_name = getattr(abstract_api, '_client_name', client_name or self._client_name)

        if abstract_api:
            self._headers = getattr(abstract_api, '_headers', None)
        else:
            self._headers = {
                'Content-Type': 'application/json',
                'Accept': 'text/plain, application/json',
                'User-Agent': f"{self._client_name} {__version__}"
            }

            if headers:
                self._headers.update({self._capitalize_header(k): v for k, v in headers.items()})

        self._max_tries = getattr(abstract_api, '_max_tries', max_tries)

    def _get_max_tries(self):
        return self._max_tries

    def get_root_url(self):
        return self._root_url

    @property
    def user_agent(self):
        return self._headers.get('User-Agent') or self._client_name

    @staticmethod
    def _capitalize_header(name: str) -> str:
        return "-".join([n.capitalize() for n in name.split('-')])

    ###############################################################################################################
    # Basic requests
    ###############################################################################################################

    def get_binary(self, url: str, accept: str = None) -> Iterator[bytes]:
        """
        Implementation of GET for binary data.

        :param url: Url to use
        :param accept: Mimetype for accept. Will be set to */* if not given.
        :return: Yields over raw chunks of the response payload.
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _get_binary() -> Iterator[bytes]:
            with self._session.get(url,
                                   headers=self._get_headers(
                                       {"Content-Type": None, "Accept": (accept or "*/*")}
                                   ),
                                   verify=self.ssl_config.get_verify(),
                                   cert=self.ssl_config.get_cert(),
                                   timeout=self._timeout,
                                   stream=True,
                                   proxies=self._get_proxies()) as res:
                self._log_communication(res, response_body=False)
                self._check_response(res)
                self._check_status_error(res)

                yield from res.iter_content(chunk_size=65536)

        yield from _get_binary()

    def post_binary(self,
                    url: str,
                    data: Any,
                    content_type: str = None,
                    expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of POST for binary data.

        :param url: Url to use
        :param data: The payload to POST. This can be anything 'requests.post(data=...)' supports.
        :param content_type: The content type of the data. Defaults to "application/octet-stream" internally if unset.
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _post_binary() -> Any:
            res = self._session.post(url,
                                     data=data,
                                     headers=self._get_headers(
                                         {"Content-Type": (content_type or "application/octet-stream")}
                                     ),
                                     verify=self.ssl_config.get_verify(),
                                     cert=self.ssl_config.get_cert(),
                                     timeout=self._timeout,
                                     proxies=self._get_proxies())
            self._log_communication(res, request_body=False)
            return self._parse_response(res, expected_media_type)

        return _post_binary()

    def put_binary(self,
                   url: str,
                   data: Any,
                   content_type: str = None,
                   expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of PUT for binary data.

        :param url: Url to use
        :param data: The payload to PUT. This can be anything 'requests.put(data=...)' supports.
        :param content_type: The content type of the data. Defaults to "application/octet-stream" internally if unset.
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _put_binary() -> Any:
            res = self._session.put(url,
                                    data=data,
                                    headers=self._get_headers(
                                        {"Content-Type": (content_type or "application/octet-stream")}
                                    ),
                                    verify=self.ssl_config.get_verify(),
                                    cert=self.ssl_config.get_cert(),
                                    timeout=self._timeout,
                                    proxies=self._get_proxies())
            self._log_communication(res, request_body=False)
            return self._parse_response(res, expected_media_type)

        return _put_binary()

    def get(self,
            url: str,
            expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of GET

        :param url: Url to use
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _get() -> Any:
            res = self._session.get(url,
                                    headers=self._get_headers({"Content-Type": None}),
                                    verify=self.ssl_config.get_verify(),
                                    cert=self.ssl_config.get_cert(),
                                    timeout=self._timeout,
                                    proxies=self._get_proxies())
            self._log_communication(res)
            return self._parse_response(res, expected_media_type)

        return _get()

    def post(self,
             url: str,
             data: Any,
             expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of POST

        :param url: Url to use
        :param data: The payload to POST
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _post() -> Any:
            res = self._session.post(url,
                                     json=data,
                                     headers=self._get_headers(),
                                     verify=self.ssl_config.get_verify(),
                                     cert=self.ssl_config.get_cert(),
                                     timeout=self._timeout,
                                     proxies=self._get_proxies())
            self._log_communication(res)
            return self._parse_response(res, expected_media_type)

        return _post()

    def put(self,
            url: str,
            data: Any,
            expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of PUT

        :param url: Url to use
        :param data: The payload to PUT
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _put() -> Any:
            res = self._session.put(url,
                                    json=data,
                                    headers=self._get_headers(),
                                    verify=self.ssl_config.get_verify(),
                                    cert=self.ssl_config.get_cert(),
                                    timeout=self._timeout,
                                    proxies=self._get_proxies())
            self._log_communication(res)
            return self._parse_response(res, expected_media_type)

        return _put()

    def patch(self,
              url: str,
              data: Any,
              expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of PATCH

        :param url: Url to use
        :param data: The payload to PUT
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _patch() -> Any:
            res = self._session.patch(url,
                                      json=data,
                                      headers=self._get_headers(),
                                      verify=self.ssl_config.get_verify(),
                                      cert=self.ssl_config.get_cert(),
                                      timeout=self._timeout,
                                      proxies=self._get_proxies())
            self._log_communication(res)
            return self._parse_response(res, expected_media_type)

        return _patch()

    def delete(self,
               url: str,
               expected_media_type: str = 'application/json') -> Any:
        """
        Implementation of DELETE

        :param url: Url to use
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The payload of the response
        """

        @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS, max_tries=self._get_max_tries)
        def _delete() -> Any:
            res = self._session.delete(url,
                                       headers=self._get_headers({"Content-Type": None}),
                                       verify=self.ssl_config.get_verify(),
                                       cert=self.ssl_config.get_cert(),
                                       timeout=self._timeout,
                                       proxies=self._get_proxies())
            self._log_communication(res)
            return self._parse_response(res, expected_media_type)

        return _delete()

    ###############################################################################################################
    # Tool methods for requests
    ###############################################################################################################

    def _get_proxies(self) -> dict:
        """
        Create a copy of proxies if they exist or return None

        :return: copy of self._proxies or None
        """
        return self._proxies.copy() if self._proxies else None

    def _get_headers(self, override: dict = None) -> dict:
        """
        Create a header dict for requests. Uses abstract method *self._handle_token()*.

        :param override: Dict of headers that override the internal headers. If a header key is set to value None,
               it will be removed from the headers.
        :return: A dict containing header values for requests.
        """
        headers = self._headers.copy()

        if isinstance(override, dict):
            headers.update({self._capitalize_header(k): v for k, v in override.items()})
            headers = {k: v for k, v in headers.items() if v is not None}

        token = self._handle_token()
        if token:
            headers['Authorization'] = "Bearer " + token

        return headers

    @staticmethod
    def _bool_to_external_str(value: Any) -> Optional[str]:
        """
        Translate bool to string values "true" and "false" if value is a bool.

        :param value: The value to check.
        :return: String representation of the value or None.
        """
        if value is None:
            return None
        if isinstance(value, bool):
            return "true" if value else "false"
        return str(value)

    @staticmethod
    def _get_query_part(params: dict) -> str:
        """
        Create the query part of an url. Keys in *params* whose values are set to None are removed.

        :param params: A dict of params to use for the query.
        :return: The query part of an url with a leading '?', or an empty string when query is empty.
        """
        params_cleaned = {k: AbstractAPI._bool_to_external_str(v) for k, v in params.items() if v is not None}
        return ('?' + urlencode(params_cleaned, quote_via=quote, safe="/,")) if params_cleaned else ""

    def _parse_response(self,
                        res: requests.Response,
                        expected_media_type: str = 'application/json') -> Any:
        """
        Parse the response of the backend.

        :param res: The result payload
        :param expected_media_type: The expected media type. Default is 'application/json'. If this is set to '*' or
               '*/*', any media_type is accepted.
        :return: The result payload. A json type when the result media_type within Content-Type is 'application/json'
                 (usually a dict), a str otherwise.
        :raises RequestException: On HTTP errors.
        :raises WrongContentTypeError: When the Media-Type of the Content-Type of the Response is not
                *expected_media_type*.
        """
        try:
            self._check_response(res)
            self._check_status_error(res)
            if expected_media_type not in ['*', '*/*']:
                AbstractAPI._check_content_type(res, expected_media_type)
            if expected_media_type.lower() == 'application/json':
                return res.json()
            else:
                return str(res.text)
        except (json.JSONDecodeError, ValueError):
            return {"error": {"message": res.text, "code": 999}}

    def _log_communication(self, res: requests.Response, request_body: bool = True, response_body: bool = True) -> None:
        """
        Log communication that flows across requests' methods. Contains options to disable logging of the body portions
        of the requests to avoid dumping large amounts of data and breaking streaming of results.

        This log does not log an exact binary representation of the transmitted bodies but uses the encoding defined
        in *res* to print it as strings.

        Authorization and cookie headers will be obscured by only displaying the last six characters of their values.

        :param res: The response of a request. Also contains the request.
        :param request_body: Option to disable the logging of the request_body.
        :param response_body: Option to disable the logging of the response_body.
        :return:
        """

        def _log_headers(headers) -> str:
            result: str = ""
            for k, v in headers.items():
                cap_key: str = self._capitalize_header(k)
                if cap_key == "Authorization" or cap_key.find("Cookie") != -1:
                    v = f"{v[:6]}[{len(v) - 12} characters hidden]{v[-6:]}"
                result += f"{k}: {v}\n"
            return result

        def _body_str(body: Union[str, bytes], encoding: str) -> str:
            if body is None:
                return ""
            if isinstance(body, bytes):
                body = str(body, encoding or 'utf8')
            return body

        ok = self._check_response_ok(res)

        if (not ok and self._log_communication_on_error) or logger.isEnabledFor(logging.DEBUG):
            log_message = f'''
################ request ################
{res.request.method} {res.request.url}
{_log_headers(res.request.headers)}
{_body_str(res.request.body, res.encoding) if request_body else "(body hidden)"}
################ response ################
{res.status_code} {res.reason} {res.url}
{_log_headers(res.headers)}
{_body_str(res.text, res.encoding) if response_body else "(body hidden)"}
'''

            if not ok:
                logger.error(log_message)
            else:
                logger.debug(log_message)

    def _check_status_error(self, res: requests.Response) -> None:
        """
        Catch exceptions and rethrow them with additional information returned by the error response body.

        :param res: The response
        :raises requests.exceptions.HTTPError: When an HTTPError occurred.
        """
        try:
            if self._raise_exceptions:
                res.raise_for_status()
                if res.status_code > 600:
                    raise requests.exceptions.HTTPError(
                        u'%s Illegal return code: %s for url: %s' % (res.status_code, res.reason, res.url),
                        response=res)

        except requests.exceptions.HTTPError as err:
            http_error_msg = str(err.args[0])

            if res.content:
                try:
                    AbstractAPI._check_content_type(res, 'application/json')
                    http_error_msg += ": " + self._get_error_message(res.json())
                except (json.JSONDecodeError, KeyError, WrongContentTypeError):
                    if '_TOKEN' not in res.text:
                        http_error_msg += ": " + str(res.text)

            raise requests.exceptions.HTTPError(http_error_msg, response=err.response) from err

    def _get_error_message(self, json_result: dict) -> str:
        """
        Extract error message from {"error": { "message": "(errormessage)" }} or {"error":"(errormessage)" } if
        possible. Return the complete JSON as str if not.

        Child API classes should overwrite this method if their error messages differ.

        :param json_result: The json dict containing the error message.
        :return: The extracted error message.
        """
        if 'error' in json_result:
            error_block = json_result.get('error')
            if isinstance(error_block, dict):
                error_message = error_block.get('message')
                return error_message if isinstance(error_message, str) else json.dumps(error_block)
            if isinstance(error_block, str):
                return error_block

        return json.dumps(json_result)

    @staticmethod
    def _check_content_type(res: requests.Response, expected_media_type: str) -> None:
        """
        Raise an Exception if the Content-Type header of the response does not contain the *expected_media_type*.
        Compares only the media-type portion of the header (by splitting at ';').

        :param res: The response object
        :param expected_media_type: The expected Media-Type.
        :raise WrongContentTypeError: When the Media-Type of the Content-Type is not *expected_media_type* or
               the header is missing completely.
        """
        content_type = res.headers.get('Content-Type')
        if not content_type:
            raise WrongContentTypeError("Response has no Content-Type.")

        media_type = content_type.lower().split(';')[0]

        if media_type != expected_media_type.lower():
            raise WrongContentTypeError(
                f"Expected media-type '{expected_media_type}' in Content-Type header, but got '{media_type}'."
            )

    def _check_response_ok(self, res: requests.Response) -> bool:
        """
        Do not rely on res.ok. Everything not between 200 and 399 is an error.

        Child API classes should overwrite this method if other custom and valid status codes might occur there.

        :param res: The response object
        :return: True on good response, false otherwise.
        """
        return 200 <= res.status_code < 400

    ###############################################################################################################
    # Response and token handling
    # Child classes have to override those classes for special handling like token and header handling.
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        Root method. No response checking here.

        :param res: The result payload
        """
        return

    def _handle_token(self) -> Optional[str]:
        """
        Just return None, therefore a header without Authorization
        will be created in *self._get_headers()*.

        Does *not* try to obtain or refresh a token.

        :return: Always None here, derived classes should return a token string.
        """
        return None


###################################################################################################################
# ConnectionHandler class
###################################################################################################################

class GraphConnectionHandler(AbstractAPI):
    """
    Contains information about a Graph Connection. This class also handles resolving the current api endpoints.
    Also creates the requests.Session which will be shared among the API Modules using this connection.
    """

    _pool_maxsize = 10
    """Default pool_maxsize for requests.adapters.HTTPAdapter."""

    _pool_block = False

    _version_info: dict = None

    _lock: threading.RLock
    """Reentrant mutex for thread safety"""

    def __init__(self,
                 root_url: str = None,
                 custom_endpoints: dict = None,
                 pool_maxsize: int = None,
                 pool_block: bool = None,
                 connection_handler=None,
                 *args,
                 **kwargs):
        """
        Constructor

        Example for custom_endpoints (see params below):

        ::

           {
               "graph": "/api/graph/7.2",
               "auth": "/api/auth/6.2",
               "action-ws": ("/api/action-ws/1.0", "action-1.0.0")
           }

        This object creates the *requests.Session* and *requests.adapters.HTTPAdapter* for this *root_url*. The
        *pool_maxsize* of such a session can be set via the parameter in the constructor. When a TokenApiHandler is
        shared between different API objects (like HiroGraph, HiroApp, etc.), this session and its pool are also
        shared.

        See parent :class:`AbstractAPI` for a description of all remaining parameters.

        :param root_url: Root url for HIRO, like https://core.arago.co.
        :param custom_endpoints: Optional map of {name:endpoint_path, ...} that overrides or adds to the endpoints taken
               from /api/version. Example see above.
        :param pool_maxsize: Size of a connection pool for a single connection. See requests.adapters.HTTPAdapter.
               Default is 10. *pool_maxsize* is ignored when *session* is set.
        :param pool_block: Block any connections that exceed the pool_maxsize. Default is False: Allow more connections,
               but do not cache them. See requests.adapters.HTTPAdapter. *pool_block* is ignored when *session* is set.
        :param connection_handler: Copy parameters from this already existing connection handler. Overrides all other
               parameters.
        :param args: Unnamed parameter passthrough for parent class.
        :param kwargs: Named parameter passthrough for parent class.
        """
        self._lock = threading.RLock()

        root_url = getattr(connection_handler, '_root_url', root_url)
        session = getattr(connection_handler, '_session', None)

        if not root_url:
            raise ValueError("'root_url' must not be empty.")

        if not session:
            adapter = requests.adapters.HTTPAdapter(
                pool_maxsize=pool_maxsize or self._pool_maxsize,
                pool_connections=1,
                pool_block=pool_block or self._pool_block
            )
            session = requests.Session()
            session.mount(root_url, adapter)

        super().__init__(
            root_url=root_url,
            session=session,
            abstract_api=connection_handler,
            *args,
            **kwargs
        )

        self.custom_endpoints = getattr(connection_handler, '_custom_endpoints', custom_endpoints)
        self._version_info = getattr(connection_handler, '_version_info', None)

        self.get_version()

    @staticmethod
    def _remove_slash(endpoint: str) -> str:
        return endpoint[:-1] if endpoint[-1] == '/' else endpoint

    ###############################################################################################################
    # Public methods
    ###############################################################################################################

    def get_api_endpoint_of(self, api_name: str) -> str:
        """
        Determines endpoints of the API names.
        Loads and caches the current API information if necessary.

        :param api_name: Name of the HIRO API
        :return: Full url of endpoint for this API
        """

        if self.custom_endpoints:
            endpoint = self.custom_endpoints.get(api_name)
            if endpoint:
                return self._remove_slash(self._root_url + endpoint)

        api_entry: dict = self.get_version().get(api_name)

        if not api_entry:
            raise ValueError("No API named '{}' found.".format(api_name))

        return self._remove_slash(self._root_url + api_entry.get('endpoint'))

    def get_websocket_config(self, api_name: str) -> Tuple[
        str,
        str,
        Optional[str],
        Optional[int],
        Optional[dict]
    ]:
        """
        Determines endpoints for websockets of the API names.
        Loads and caches the current API information if necessary.
        If proxies have been given, the key of the proxy picked needs to be "ws" or "wss" respectively.

        :param api_name: Name of the HIRO API for websockets
        :return: Tuple of full url of websocket for this API, its protocol, its proxy_host, its proxy port and proxy
                 auth (if any).
        """

        def _get_proxy_info(_url: str) -> Tuple[Optional[str], Optional[int], Optional[dict]]:
            proxies = self._get_proxies()
            if not proxies:
                return None, None, None

            url_parts = urllib.parse.urlparse(_url)
            proxy_url = proxies.get(url_parts.scheme)

            if not proxy_url:
                return None, None, None

            proxy_url_parts = urllib.parse.urlparse(proxy_url)

            proxy_auth: dict = {
                proxy_url_parts.username: proxy_url_parts.password
            } if proxy_url_parts.username else None

            return proxy_url_parts.hostname, proxy_url_parts.port, proxy_auth

        def _construct_result(_endpoint: str, _protocol: str) -> Tuple[
            str,
            str,
            Optional[str],
            Optional[str],
            Optional[dict]
        ]:
            _url: str = self._root_url.lower().replace('https://', 'wss://').replace('http://', 'ws://')
            _proxy, _proxy_port, _proxy_auth = _get_proxy_info(_url)
            return self._remove_slash(_url + _endpoint), _protocol, _proxy, _proxy_port, _proxy_auth

        if self.custom_endpoints:
            value = self.custom_endpoints.get(api_name)
            if isinstance(value, tuple):
                endpoint, protocol = value
                if endpoint:
                    return _construct_result(endpoint, protocol)

        api_entry: dict = self.get_version().get(api_name)

        if not api_entry:
            raise ValueError("No WS API named '{}' found.".format(api_name))

        return _construct_result(api_entry.get('endpoint'), api_entry.get('protocol'))

    ###############################################################################################################
    # REST API operations
    ###############################################################################################################

    def get_version(self, force_update: bool = False) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/api/version'`

        :param force_update: Force updating the internal cache with version_info via API request.
        :return: The result payload
        """
        with self._lock:
            if not self._version_info or force_update:
                url = self._root_url + '/api/version'
                self._version_info = self.get(url)

            return self._version_info


###################################################################################################################
# TokenApiHandler classes
###################################################################################################################

class AbstractTokenApiHandler(GraphConnectionHandler):
    """
    Root class for all TokenApiHandler classes. This adds token handling.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor

        See parent :class:`GraphConnectionHandler` for a full description
        of all remaining parameters.

        :param args: Unnamed parameter passthrough for parent class.
        :param kwargs: Named parameter passthrough for parent class.
        """
        super().__init__(*args, **kwargs)

    ###############################################################################################################
    # Token handling
    ###############################################################################################################

    @property
    def token(self) -> str:
        """
        Return the current token.
        """
        raise RuntimeError('Cannot use property of this abstract class.')

    def decode_token(self) -> dict:
        """
        Return a dict with the decoded token payload. This payload contains detailed information about what this token
        has access to.
        :return: The dict with the decoded token payload.
        """
        base64_payload: list = self.token.split('.')
        if len(base64_payload) == 1:
            raise AuthenticationTokenError("Token is missing base64 payload")

        payload = base64_payload[1] + '=' * (4 - len(base64_payload[1]) % 4)

        json_payload = base64.urlsafe_b64decode(payload)

        return json.loads(json_payload)

    @abstractmethod
    def refresh_token(self) -> None:
        """
        Refresh the current token.
        """
        raise RuntimeError('Cannot use method of this abstract class.')

    @abstractmethod
    def refresh_time(self) -> Optional[int]:
        """
        Calculate the time after which the token should be refreshed in milliseconds.

        :return: The timestamp in ms after which the token shall be refreshed or None if the token cannot be refreshed
                 on its own.
        """
        raise RuntimeError('Cannot use method of this abstract class.')


class FixedTokenApiHandler(AbstractTokenApiHandler):
    """
    TokenApiHandler for a fixed token.
    """

    _token: str

    def __init__(self, token: str = None, *args, **kwargs):
        """
        Constructor

        See parent :class:`AbstractTokenApiHandler` for a full description
        of all remaining parameters.

        :param token: The fixed token for the HTTP requests.
        :param args: Unnamed parameter passthrough for parent class. See :class:`AbstractTokenApiHandler`.
        :param kwargs: Named parameter passthrough for parent class. See :class:`AbstractTokenApiHandler`.
        """
        super().__init__(*args, **kwargs)

        self._token = token

    @property
    def token(self) -> str:
        return self._token

    def refresh_token(self) -> None:
        raise FixedTokenError('Token is invalid and cannot be changed because it has been given externally.')

    def refresh_time(self) -> Optional[int]:
        """

        :return: Always none
        """
        return None


class EnvironmentTokenApiHandler(AbstractTokenApiHandler):
    """
    TokenApiHandler for a fixed token given as an environment variable.
    """

    _env_var: str

    def __init__(self, env_var: str = 'HIRO_TOKEN', *args, **kwargs):
        """
        Constructor

        See parent :class:`AbstractTokenApiHandler` for a full description
        of all remaining parameters.

        :param env_var: Name of the environment variable with the token.
        :param args: Unnamed parameter passthrough for parent class.
        :param kwargs: Named parameter passthrough for parent class. 
        """
        super().__init__(*args, **kwargs)

        self._env_var = env_var

    @property
    def token(self) -> str:
        return os.environ[self._env_var]

    def refresh_token(self) -> None:
        raise FixedTokenError(
            "Token is invalid and cannot be changed because it has been given as environment variable '{}'"
            " externally.".format(self._env_var))

    def refresh_time(self) -> Optional[int]:
        """

        :return: Always none
        """
        return None


class TokenInfo:
    """
    This class stores token information from the auth api.
    """

    token: str = None
    """ The token string """
    expires_at = -1
    """ Token expiration in ms since epoch"""
    refresh_token: str = None
    """ The refresh token to use - if any."""
    last_update = 0
    """ Timestamp of when the token has been fetched in ms."""
    refresh_offset = 5000
    """ Milliseconds of offset for token expiry """

    def __init__(self, token: str = None, refresh_token: str = None, expires_at: int = -1, refresh_offset: int = 5000):
        """
        Constructor

        :param token: The token string
        :param refresh_token: A refresh token
        :param expires_at: Token expiration in ms since epoch
        :param refresh_offset: Offset in milliseconds that will be subtracted from the expiry time so a token will be
                               refreshed in time. Default is 5 seconds.
        """
        self.token = token
        self.expires_at = expires_at
        self.refresh_token = refresh_token
        self.last_update = self.get_epoch_millis() if token else 0
        self.refresh_offset = refresh_offset

    @staticmethod
    def get_epoch_millis() -> int:
        """
        Get timestamp
        :return: Current epoch in milliseconds
        """
        return int(time.time_ns() / 1000000)

    def parse_token_result(self, res: dict, what: str) -> None:
        """
        Parse the result payload and extract token information.

        :param res: The result payload from the backend.
        :param what: What token command has been issued (for error messages).
        :raises TokenUnauthorizedError: When the token request returned error 401. This usually means, that this token
                has expired.
        :raises AuthenticationTokenError: When the token request returned any other error.
        """
        if 'error' in res:
            message: str = '{}: {}'.format(what, res['error'].get('message'))
            code: int = int(res['error'].get('code'))

            if code == 401:
                raise TokenUnauthorizedError(message, code)
            else:
                raise AuthenticationTokenError(message, code)

        self.last_update = self.get_epoch_millis()

        self.token = res.get('_TOKEN')

        expires_at = res.get('expires-at')
        if expires_at:
            self.expires_at = int(expires_at)
        else:
            expires_in = res.get('expires_in')
            if expires_in:
                self.expires_at = self.last_update + int(expires_in) * 1000

        refresh_token = res.get('refresh_token')
        if refresh_token:
            self.refresh_token = refresh_token

    def expired(self) -> bool:
        """
        Check token expiration

        :return: True when the token has been expired *(expires_at - refresh_offset) <= get_epoch_mills()*
        """
        return self.refresh_time() <= self.get_epoch_millis()

    def fresh(self, span: int = 30000) -> bool:
        """
        Check, whether the last token fetched is younger than span ms.

        :param span: Timespan in ms in which a token is considered fresh. Default is 30 sec (30000ms).
        :return: True when the last update was less than span ms.
        """

        return (self.get_epoch_millis() - self.last_update) < span

    def refresh_time(self) -> Optional[int]:
        """
        Calculate the time after which the token should be refreshed in milliseconds.

        :return: expires_at - refresh_offset (in ms) or None if refresh is not possible.
        """
        return self.expires_at - self.refresh_offset if self.expires_at > 0 else None


class PasswordAuthTokenApiHandler(AbstractTokenApiHandler):
    """
    API Tokens will be fetched using this class. It does not handle any automatic token fetching, refresh or token
    expiry. This has to be checked and triggered by the *caller*.

    The methods of this class are thread-safe, so it can be shared between several HIRO objects.

    It is built this way to avoid endless calling loops when resolving tokens.
    """

    _token_info: TokenInfo = None
    """Contains all token information"""

    _lock: threading.RLock
    """Reentrant mutex for thread safety"""

    _username: str
    _password: str
    _client_id: str
    _client_secret: str

    _secure_logging: bool = True

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 client_id: str = None,
                 client_secret: str = None,
                 secure_logging: bool = True,
                 *args, **kwargs):
        """
        Constructor

        See parent :class:`AbstractTokenApiHandler` for a full description
        of all remaining parameters.

        :param username: Username for authentication
        :param password: Password for authentication
        :param client_id: OAuth client_id for authentication
        :param client_secret: OAuth client_secret for authentication
        :param secure_logging: If this is enabled, payloads that might contain sensitive information are not logged.
        :param args: Unnamed parameter passthrough for parent class. 
        :param kwargs: Named parameter passthrough for parent class. 
        """
        super().__init__(*args, **kwargs)

        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret

        self._secure_logging = secure_logging

        self._token_info = TokenInfo()
        self._lock = threading.RLock()

    @property
    def endpoint(self):
        return self.get_api_endpoint_of('auth')

    @property
    def token(self) -> str:
        """Get the token. Get or refresh it if necessary."""
        with self._lock:
            if not self._token_info.token:
                self.get_token()
            elif self._token_info.expired():
                self.refresh_token()

            return self._token_info.token

    def _log_communication(self, res: requests.Response, request_body: bool = True, response_body: bool = True) -> None:
        """
        Logging under a secure aspect. Hides sensitive information unless *self._secure_logging* is set to False.

        :param res: The response of a request. Also contains the request.
        :param request_body: Option to disable the logging of the request_body. If set to True, will only remain True
               internally when *self._secure_logging* is set to False.
        :param response_body: Option to disable the logging of the response_body.  If set to True, will only remain True
               internally when *self._secure_logging* is set to False or *res.status_code* != 200.
        """
        log_request_body = not self._secure_logging and request_body is True
        log_response_body = (res.status_code != 200 or not self._secure_logging) and response_body is True

        super()._log_communication(res, request_body=log_request_body, response_body=log_response_body)

    def get_token(self) -> None:
        """
        Construct a request to obtain a new token. API self._endpoint + '/app'

        :raises AuthenticationTokenError: When no auth_endpoint is set.
        """
        with self._lock:
            if not self.endpoint:
                raise AuthenticationTokenError(
                    'Token is invalid and endpoint (auth_endpoint) for obtaining is not set.')

            if not self._username or not self._password or not self._client_id or not self._client_secret:
                msg = ""
                if not self._username:
                    msg += "'username'"
                if not self._password:
                    msg += (", " if msg else "") + "'password'"
                if not self._client_id:
                    msg += (", " if msg else "") + "'client_id'"
                if not self._client_secret:
                    msg += (", " if msg else "") + "'client_secret'"
                raise AuthenticationTokenError(
                    "{} is missing required parameter(s) {}.".format(self.__class__.__name__, msg))

            url = self.endpoint + '/app'
            data = {
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "username": self._username,
                "password": self._password
            }

            res = self.post(url, data)
            self._token_info.parse_token_result(res, "{}.get_token".format(self.__class__.__name__))

    def refresh_token(self) -> None:
        """
        Construct a request to refresh an existing token. API self._endpoint + '/refresh'.
        Does not refresh tokens that are younger than 30 sec to avoid refresh storms on parallel connections.

        :raises AuthenticationTokenError: When no auth_endpoint is set.
        """
        with self._lock:
            if not self.endpoint:
                raise AuthenticationTokenError(
                    'Token is invalid and endpoint (auth_endpoint) for refresh is not set.')

            if self._token_info.fresh():
                return

            if not self._token_info.refresh_token:
                self.get_token()
                return

            url = self.endpoint + '/refresh'
            data = {
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "refresh_token": self._token_info.refresh_token
            }

            try:
                res = self.post(url, data)
                self._token_info.parse_token_result(res, "{}.refresh_token".format(self.__class__.__name__))
            except AuthenticationTokenError:
                self.get_token()

    def refresh_time(self) -> Optional[int]:
        """
        Calculate refresh time.

        :return: Timestamp after which the token becomes invalid.
        """
        return self._token_info.refresh_time()

    ###############################################################################################################
    # Response and token handling
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        Response checking. When a refresh_token is present and status_code is 401, raise TokenUnauthorizedError.
        This can happen when a refresh-token is not valid anymore.

        :param res: The result payload
        :raise TokenUnauthorizedError: Raised to trigger a retry via self.get_token() in self.refresh_token().
        """
        if res.status_code == 401 and self._token_info.refresh_token:
            raise TokenUnauthorizedError(str(res.text), 401)

    def _handle_token(self) -> Optional[str]:
        """
        Just return None, therefore a header without Authorization
        will be created in *self._get_headers()*.

        Does *not* try to obtain or refresh a token.

        :return: *token* given.
        """
        return None


###################################################################################################################
# Root class for different API groups
###################################################################################################################

class AuthenticatedAPIHandler(AbstractAPI):
    """
    Python implementation for accessing a REST API with authentication.
    """

    _api_handler: AbstractTokenApiHandler
    _api_name: str

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 api_name: str):
        """
        Constructor

        :param api_name: Name of the API to use.
        :param api_handler: External API handler.
        """
        if not api_handler or not api_name:
            raise ValueError("Cannot authenticate against HIRO without *api_handler* and *api_name*.")

        super().__init__(abstract_api=api_handler)

        self._api_handler = api_handler
        self._api_name = api_name

    @property
    def endpoint(self) -> str:
        return self._api_handler.get_api_endpoint_of(self._api_name)

    ###############################################################################################################
    # Response and token handling
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        Response checking. Tries to refresh the token on status_code 401, then raises RequestException to try
        again using backoff.

        :param res: The result payload
        :raises requests.exceptions.RequestException: When an error 401 occurred and the token has been refreshed.
        """
        if res.status_code == 401:
            self._api_handler.refresh_token()

            # Raise this exception to trigger retry with backoff
            raise requests.exceptions.RequestException

    def _handle_token(self) -> Optional[str]:
        """
        Try to return a valid token by obtaining or refreshing it.

        :return: A valid token.
        """
        return self._api_handler.token


###################################################################################################################
# Exceptions
###################################################################################################################

class AuthenticationTokenError(Exception):
    """
    Class for unrecoverable failures with access tokens.
    Contains a message and an optional message code. If the code is None, no code will be printed in __str__().
    """
    message: str
    code: int

    def __init__(self, message: str, code: int = None):
        self.message = message
        self.code = code

    def __str__(self) -> str:
        if self.code is None:
            return "{}: {}".format(self.__class__.__name__, self.message)
        else:
            return "{}: {} ({})".format(self.__class__.__name__, self.message, self.code)


class TokenUnauthorizedError(AuthenticationTokenError):
    """
    Child of *AuthenticationTokenErrors*. Used when tokens expire with error 401.
    """
    pass


class FixedTokenError(AuthenticationTokenError):
    """
    Child of *AuthenticationTokenErrors*. Used when tokens are fixed and cannot be refreshed.
    """
    pass


class WrongContentTypeError(Exception):
    """
    When the Content-Type of the result is unexpected, i.e. 'application/json' was expected, but something else got
    returned.
    """
    pass



"""
Package which contains the classes to communicate with HIRO Graph.
"""
import site
from os import path

from hiro_graph_client.appclient import HiroApp
from hiro_graph_client.authclient import HiroAuth
from hiro_graph_client.authzclient import HiroAuthz
from hiro_graph_client.client import HiroGraph
from hiro_graph_client.clientlib import AbstractTokenApiHandler, GraphConnectionHandler, AuthenticationTokenError, \
    FixedTokenError, TokenUnauthorizedError, PasswordAuthTokenApiHandler, FixedTokenApiHandler, \
    EnvironmentTokenApiHandler, SSLConfig
from hiro_graph_client.iamclient import HiroIam
from hiro_graph_client.kiclient import HiroKi
from hiro_graph_client.variablesclient import HiroVariables
from hiro_graph_client.version import __version__

this_directory = path.abspath(path.dirname(__file__))

__all__ = [
    'HiroGraph', 'HiroAuth', 'HiroApp', 'HiroIam', 'HiroKi', 'HiroAuthz', 'HiroVariables', 'GraphConnectionHandler',
    'AbstractTokenApiHandler', 'PasswordAuthTokenApiHandler', 'FixedTokenApiHandler', 'EnvironmentTokenApiHandler',
    'AuthenticationTokenError', 'FixedTokenError', 'TokenUnauthorizedError', '__version__',
    'SSLConfig'
]

site.addsitedir(this_directory)



#!/usr/bin/env python3
import datetime
from typing import Any, Iterator, Union, List, Dict
from urllib.parse import quote_plus

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroGraph(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO HIRO REST API.
    See https://core.arago.co/help/specs/?url=definitions/graph.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='graph',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations
    ###############################################################################################################

    def query(self,
              query: str,
              fields: str = None,
              limit=-1,
              offset=0,
              order: str = None,
              meta: bool = None,
              count: bool = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Search/post_query_vertices

        :param query: The actual query. e.g. ogit\\\\/_type: ogit\\\\/Question for vertices.
        :param fields: the comma separated list of fields to return
        :param limit: limit of entries to return
        :param offset: offset where to start returning entries
        :param order: order by a field asc|desc, e.g. ogit/name desc
        :param meta: List detailed metainformations in result payload
        :param count: Just return the number of found items. Result payload is like
               {"items":[&lt;number of items found as int&gt;]}.
        :return: Result payload
        """
        url = self.endpoint + '/query/vertices'

        data = {"query": str(query)}
        if fields:
            data['fields'] = quote_plus(fields.replace(" ", ""), safe="/,")
        if limit is not None:
            data['limit'] = limit
        if offset:
            data['offset'] = offset
        if order:
            data['order'] = order
        if meta is not None:
            data['listMeta'] = meta
        if count is not None:
            data['count'] = count
        return self.post(url, data)

    def query_gremlin(self,
                      query: str,
                      root: str,
                      fields: str = None,
                      include_deleted: bool = None,
                      meta: bool = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Search/post_query_gremlin

        :param query: The actual query. e.g. outE().inV() for gremlin.
        :param root: ogit/_id of the root node where the gremlin query starts.
        :param fields: the comma separated list of fields to return
        :param include_deleted: Include deleted values.
        :param meta: List detailed metainformations in result payload
        :return: Result payload
        """
        url = self.endpoint + '/query/gremlin'

        data = {"query": str(query),
                "root": root}
        if fields:
            data['fields'] = quote_plus(fields.replace(" ", ""), safe="/,")
        if include_deleted is not None:
            data['include_deleted'] = include_deleted
        if meta is not None:
            data['listMeta'] = meta
        return self.post(url, data)

    def create_node(self, data: dict, obj_type: str, return_id=False) -> Union[dict, str]:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Entity/post_new__type_

        :param data: Payload for the new node/vertex
        :param obj_type: ogit/_type of the new node/vertex
        :param return_id: Return only the ogit/_id as string. Default is False to return everything as dict.
        :return: The result payload
        """
        url = self.endpoint + '/new/' + quote_plus(obj_type)
        res = self.post(url, data)
        return res['ogit/_id'] if return_id and 'error' not in res else res

    def update_node(self, node_id: str, data: dict) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Entity/post__id_

        :param data: Payload for the node/vertex
        :param node_id: ogit/_id of the node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id)
        return self.post(url, data)

    def delete_node(self, node_id: str) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Entity/delete__id_

        :param node_id: ogit/_id of the node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id)
        return self.delete(url)

    def connect_nodes(self, from_node_id: str, verb: str, to_node_id: str) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Verb/post_connect__type_

        :param from_node_id: ogit/_id of the source node/vertex
        :param verb: verb for the connection
        :param to_node_id: ogit/_id of the target node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/connect/' + quote_plus(verb)
        data = {"out": from_node_id, "in": to_node_id}
        return self.post(url, data)

    def disconnect_nodes(self, from_node_id: str, verb: str, to_node_id: str) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Verb/delete__id_

        :param from_node_id: ogit/_id of the source node/vertex
        :param verb: verb for the connection
        :param to_node_id: ogit/_id of the target node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(
            from_node_id
        ) + "$$" + quote_plus(
            verb
        ) + "$$" + quote_plus(
            to_node_id
        )
        return self.delete(url)

    def get_node(self,
                 node_id: str,
                 fields: str = None,
                 meta: bool = None,
                 include_deleted: bool = None,
                 vid: str = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Graph]_Entity/get__id_

        :param node_id: ogit/_id of the node/vertex or edge
        :param fields: Filter for fields
        :param include_deleted: allow to get if ogit/_is-deleted=true
        :param vid: get specific version of Entity matching ogit/_v-id
        :param meta: List detailed metainformations in result payload
        :return: The result payload
        """
        query = {
            "fields": fields.replace(" ", "") if fields else None,
            "listMeta": meta,
            "includeDeleted": include_deleted,
            "vid": vid
        }

        url = self.endpoint + '/' + quote_plus(node_id) + self._get_query_part(query)
        return self.get(url)

    def get_nodes(self,
                  node_ids: list,
                  fields: str = None,
                  meta: bool = None,
                  include_deleted: bool = None,
                  ) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Search/get_query_ids

        :param node_ids: list of ogit/_ids of the node/vertexes or edges
        :param fields: Filter for fields
        :param meta: List detailed metainformations in result payload
        :param include_deleted: allow to get if ogit/_is-deleted=true
        :return: The result payload
        """
        query = {
            "query": ",".join(node_ids),
            "fields": fields.replace(" ", "") if fields else None,
            "includeDeleted": include_deleted,
            "listMeta": meta
        }

        url = self.endpoint + '/query/ids' + self._get_query_part(query)
        return self.get(url)

    def get_node_by_xid(self,
                        node_id: str,
                        fields: str = None,
                        meta: bool = None,
                        include_deleted: bool = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Search/get_xid__id_

        :param node_id: ogit/_xid of the node/vertex or edge
        :param fields: Filter for fields
        :param meta: List detailed metainformations in result payload
        :param include_deleted: allow to get if ogit/_is-deleted=true
        :return: The result payload
        """
        query = {
            "fields": fields.replace(" ", "") if fields else None,
            "includeDeleted": include_deleted,
            "listMeta": meta
        }

        url = self.endpoint + '/xid/' + quote_plus(node_id) + self._get_query_part(query)
        return self.get(url)

    def get_timeseries(self,
                       node_id: str,
                       starttime: str = None,
                       endtime: str = None,
                       include_deleted: bool = None,
                       limit: int = None,
                       with_ids: str = None,
                       order: str = "asc",
                       aggregate: str = None) -> Union[List, Dict]:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Timeseries/get__id__values

        :param node_id: ogit/_id of the node containing timeseries
        :param starttime: ms since epoch.
        :param endtime: ms since epoch.
        :param aggregate: aggregate numeric values for multiple timeseries ids with same timestamp: avg|min|max|sum|none
        :param order: order by a timestamp asc|desc|none. Default is "asc" here.
        :param with_ids: list of ids to aggregate in result
        :param limit: limit of entries to return
        :param include_deleted: allow to get if ogit/_is-deleted=true
        :return: The result payload. Either a list of dict or a dict with an error message.
        """
        query = {
            "from": starttime,
            "to": endtime,
            "include_deleted": include_deleted,
            "limit": limit,
            "with": with_ids,
            "order": order,
            "aggregate": aggregate
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/values' + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        timeseries: list = res['items']
        return timeseries

    def get_timeseries_history(self,
                               node_id: str,
                               timestamp: str = None,
                               include_deleted: bool = None) -> Union[List, Dict]:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Timeseries/get__id__values_history

        :param node_id: ogit/_id of the node containing timeseries
        :param timestamp: timestamp in ms
        :param include_deleted: allow to get if ogit/_is-deleted=true
        :return: The result payload. Either a list of dict or a dict with an error message.
        """
        query = {
            "include_deleted": include_deleted,
            "timestamp": timestamp
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/values/history' + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        timeseries: list = res['items']
        return timeseries

    def query_timeseries(self,
                         starttime: str = None,
                         endtime: str = None,
                         limit: int = None,
                         order: str = "asc",
                         aggregate: str = None) -> Union[List, Dict]:
        """
        Run a query against the graph and return agragated timeseries values for timeseries vertices matching
        query result. query: Entities with matching ogit/_type:ogit/Timeseries

        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Search/get_query_values

        :param starttime: ms since epoch.
        :param endtime: ms since epoch.
        :param aggregate: aggregate numeric values for multiple timeseries ids with same timestamp: avg|min|max|sum|none
        :param order: order by a timestamp asc|desc|none. Default is "asc" here.
        :param limit: limit of entries to return
        :return: The result payload. Either a list of dict or a dict with an error message.
        """
        query = {
            "from": starttime,
            "to": endtime,
            "limit": limit,
            "order": order,
            "aggregate": aggregate
        }

        url = self.endpoint + '/query/values' + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        timeseries: list = res['items']
        return timeseries

    def post_timeseries(self,
                        node_id: str,
                        items: list,
                        synchronous: bool = True,
                        ttl: int = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Storage]_Timeseries/post__id__values

        :param synchronous: whether the operation should return synchronously. Default is True here.
        :param ttl: time to live for values to be stored in seconds (overrides /ttl in vertex).
        :param node_id: ogit/_id of the node containing timeseries
        :param items: list of timeseries values [{timestamp: (ms since epoch), value: ...},...]
        :return: The result payload
        """

        query = {
            "synchronous": synchronous,
            "ttl": ttl
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/values' + self._get_query_part(query)
        data = {"items": items}
        return self.post(url, data)

    def get_attachment(self,
                       node_id: str,
                       content_id: str = None,
                       include_deleted: bool = None) -> Iterator[bytes]:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_Blob/get__id__content

        :param node_id: Id of the attachment node
        :param content_id: Id of the content within the attachment node. Default is None.
        :param include_deleted: Whether to be able to access deleted content: Default is False
        :return: Returns generator over byte chunks from the response body payload.
        """
        query = {
            "contentId": content_id,
            "includeDeleted": include_deleted
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/content' + self._get_query_part(query)
        yield from self.get_binary(url)

    def post_attachment(self,
                        node_id: str,
                        data: Any,
                        content_type: str = None) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Storage]_Blob/post__id__content

        :param node_id: Id of the attachment node
        :param data: Data to upload in binary form. Can also be an IO object for streaming.
        :param content_type: Content-Type for *data*. Defaults to 'application/octet-stream' if left unset.
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id) + '/content'
        return self.post_binary(url, data, content_type=content_type)

    def get_history(self,
                    node_id: str,
                    ts_from: int = 0,
                    ts_to: int = datetime.datetime.now(),
                    history_type: str = 'element',
                    version: str = None,
                    vid: str = None,
                    limit=-1,
                    offset=0,
                    include_deleted: bool = None,
                    meta: bool = None
                    ) -> dict:
        """
        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Query]_History/get__id__history

        :param node_id: Id of the node
        :param ts_from: timestamp in ms where to start returning entries (default: 0)
        :param ts_to: timestamp in ms where to end returning entries (default: now)
        :param history_type: Response format:
                             full - full event,
                             element - only event body,
                             diff - diff to previous event.
                             (default: 'element')
        :param version: get entry with specific ogit/_v value
        :param vid: get specific version of Entity matching ogit/_v-id
        :param limit: limit of entries to return (default: -1).
        :param offset: offset where to start returning entries (default: 0)
        :param include_deleted: allow to get if ogit/_is-deleted=true (default: false)
        :param meta: return list type attributes with metadata (default: false)
        :return: The result payload
        """

        query = {
            "from": ts_from,
            "to": ts_to,
            "type": history_type,
            "version": version,
            "vid": vid,
            "limit": limit,
            "offset": offset,
            "includeDeleted": include_deleted,
            "listMeta": meta
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/history' + self._get_query_part(query)
        return self.get(url)

    def get_events(self,
                   ts_from: int = 0,
                   ts_to: int = datetime.datetime.now(),
                   ogit_type: str = None,
                   jfilter: str = None) -> dict:
        """
        Replays events from history

        https://core.arago.co/help/specs/?url=definitions/graph.yaml#/[Events]_History/get_events_

        :param ts_from: timestamp in ms where to start returning entries (default: 0)
        :param ts_to: timestamp in ms where to end returning entries (default: now)
        :param jfilter: jfilter string to limit matching results
        :param ogit_type: Entity or Verb ogit/_type for filtering result based on this type
        :return: The result payload
        """

        query = {
            "from": ts_from,
            "to": ts_to,
            "type": ogit_type,
            "filter": jfilter
        }

        url = self.endpoint + '/events' + self._get_query_part(query)
        return self.get(url)



from os import path

from setuptools import setup, find_packages

import timestamp
import version_by_git

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), encoding='utf-8') as fr:
    long_description = fr.read()
with open(path.join(this_directory, '../CHANGELOG.md'), encoding='utf-8') as fc:
    long_description += "\n---\n# CHANGELOG\n---\n"
    long_description += fc.read()

subversion = timestamp.make_timestamp()

name = 'hiro_graph_client'

__version__ = version_by_git.create_version_file(name)

setup(
    name=name,
    version=__version__,
    packages=find_packages(),
    python_requires='>=3.7',

    install_requires=[
        'wheel',
        'requests',
        'backoff',
        'websocket-client',
        'apscheduler'
    ],
    package_data={
        name: ['VERSION']
    },

    project_urls={
        'GitHub': 'https://github.com/arago/hiro-client-python',
        'Documentation': 'https://github.com/arago/hiro-client-python/blob/master/src/README.md',
        'Changelog': 'https://github.com/arago/hiro-client-python/blob/master/CHANGELOG.md'
    },

    author="arago GmbH",
    author_email="info@arago.co",
    maintainer="Wolfgang Hübner",
    description="Hiro Client for Graph REST API of HIRO 7",
    keywords="arago HIRO7 API REST WebSocket",
    url="https://github.com/arago/hiro-client-python",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries"
    ],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



#!/usr/bin/env python3

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroAuthz(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO Authz REST API.
    See https://core.arago.co/help/specs/?url=definitions/authz.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='authz',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations against the authz API
    ###############################################################################################################

    def entitlement(self, data: dict) -> dict:
        """
        Ask for entitlement decision

        HIRO REST query API: `POST self._endpoint + '/entitlement`

        :param data: Entitlement request data.
               See https://core.arago.co/help/specs/?url=definitions/authz.yaml#/[Authorization]_Entitlement/post_entitlement
        :return: The result payload
        """
        url = self.endpoint + '/entitlement'
        return self.post(url, data)



import json
import logging
import random
import ssl
import threading
import time
from abc import abstractmethod
from enum import Enum
from typing import List, Dict, Optional
from urllib.parse import urlencode

from websocket import WebSocketApp, ABNF, WebSocketException, setdefaulttimeout, WebSocketConnectionClosedException, \
    STATUS_NORMAL, STATUS_UNEXPECTED_CONDITION

from hiro_graph_client.clientlib import AbstractTokenApiHandler

logger = logging.getLogger(__name__)
""" The logger for this module """


class ErrorMessage:
    """
    The structure of an incoming error message
    """
    code: int
    message: str

    def __init__(self,
                 code,
                 message):
        """
        Constructor

        :param code: Numerical error code of the error
        :param message: Error message
        """
        self.code = int(code)
        self.message = str(message)

    @classmethod
    def parse(cls, message: str):
        """
        :param message: The message received from the websocket. Will be decoded here.
        :return: The new error message or None if this is not an error message.
        """
        json_message: dict = json.loads(message)
        error_message = json_message.get('error')
        if isinstance(error_message, dict):
            return cls(error_message.get('code'),
                       error_message.get('message'))
        else:
            return None

    def __str__(self):
        return json.dumps(vars(self))

    def to_dict(self) -> dict:
        return {
            "error": {
                "code": self.code,
                "message": self.message
            }
        }


class ReaderStatus(str, Enum):
    """
    The states the reader thread can be in.
    """
    NONE = 'Not started',
    STARTING = 'Starting',
    RUNNING_PRELIMINARY = 'Running preliminary (status of token unknown)',
    RUNNING = 'Running'
    RESTARTING = 'Restarting',
    DONE = 'Finished normally',
    FAILED = 'Finished because of error'


class AbstractAuthenticatedWebSocketHandler:
    """
    The basic class for all WebSockets.
    """
    _api_handler: AbstractTokenApiHandler
    _proxy_hostname: str
    _proxy_port: str
    _proxy_auth: dict

    _reconnect_delay: int

    _protocol: str
    _url: str

    _auto_reconnect: bool

    _ws: Optional[WebSocketApp] = None

    _reader_status: ReaderStatus
    """
    Tracks the status of the internal reader thread.  
    """
    _reader_guard: threading.Condition
    """
    Meant to protect the startup sequence. *
    """
    _ws_guard: threading.RLock
    """
    Protects the websocket reference for startup and restart.
    """

    _backoff_condition: threading.Condition

    _remote_exit_codes: List[int] = []
    """ A list of remote exit codes that will cause the websocket to not reconnect."""

    MAX_RETRIES = 3

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 api_name: str,
                 query_params: Dict[str, str] = None,
                 timeout: int = 5,
                 auto_reconnect: bool = True,
                 remote_exit_codes: List[int] = None):
        """
        Create the websocket

        :param api_handler: Required: The api handler for authentication.
        :param api_name: The name of the ws api.
        :param query_params: URL Query parameters for this specific websocket. Use Dict[str,str] only here,
                             i.e. set {"allscopes": "false"} instead of {"allscopes": False}.
        :param timeout: The timeout for websocket messages. Default is 5sec.
        :param auto_reconnect: Try to create a new websocket automatically when *self.send()* fails. If this is set
                               to False, a WebSocketException will be raised instead. The default is True.
        :param remote_exit_codes: A list of close codes that can come in from the remote side. If one of the codes in
                                  this list matches the remote close code, *self.run_forever()* will exit instead of
                                  trying to reconnect. The default is None -> reconnect always.
                                  See https://datatracker.ietf.org/doc/html/rfc6455#section-7.4.1
        """
        if not api_handler:
            raise ValueError('Parameter api_handler= cannot be empty.')

        if not api_name:
            raise ValueError('Parameter api_name= cannot be empty.')

        self._url, self._protocol, self._proxy_hostname, self._proxy_port, self._proxy_auth = \
            api_handler.get_websocket_config(api_name)

        if query_params:
            self._url += '?' + urlencode(query_params, safe="/,")

        self._api_handler = api_handler

        self._auto_reconnect = auto_reconnect

        self._reader_status = ReaderStatus.NONE
        self._reader_guard = threading.Condition()

        self._ws_guard = threading.RLock()
        self._backoff_condition = threading.Condition()

        if remote_exit_codes:
            self._remote_exit_codes = remote_exit_codes

        setdefaulttimeout(timeout)

        random.seed(time.time_ns())

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.stop()

    def _set_error(self, error: Exception) -> None:
        """
        Log the Exception, store it and set the status to FAILED.

        :param error: The Exception to store.
        """
        self._reader_status = ReaderStatus.FAILED
        logger.error("Reader encountered error: %s %s", error.__class__.__name__, str(error))

    def _check_message(self, ws: WebSocketApp, message: str) -> None:
        """
        Look for error 401. Try to reconnect with a new token when this is encountered.
        Set status to *ReaderStatus.RESTARTING* when a token is no longer valid on error code 401.
        Set status to *ReaderStatus.FAILED* when error 401 is received and the token was never valid.

        :param ws: WebSocketApp
        :param message: Incoming message as string
        """
        try:
            with self._reader_guard:
                error_message = ErrorMessage.parse(message)
                if error_message:
                    if error_message.code == 401:
                        if self._reader_status == ReaderStatus.RUNNING_PRELIMINARY:
                            raise WebSocketException(
                                "Received error message while token was never valid: " + str(error_message))
                        else:
                            self._api_handler.refresh_token()

                            # If we get here, the token has been refreshed successfully.
                            self._reconnect_delay = 0
                            self._reader_status = ReaderStatus.RESTARTING

                            logger.info("Refreshing token because of error: %s", str(error_message))
                            self._close(reason=f"{self._api_handler.user_agent} token refresh.")
                            return
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Received message: %s", message)
                    else:
                        logger.info("Received message of length %d", len(message))

            self.on_message(ws, message)

            with self._reader_guard:
                # If we get here, the token is valid
                if self._reader_status not in [ReaderStatus.RUNNING, ReaderStatus.FAILED]:
                    self._reader_status = ReaderStatus.RUNNING
                    self._reconnect_delay = 0

        except Exception as err:
            self._set_error(err)
            self._close(status=STATUS_UNEXPECTED_CONDITION,
                        reason="Exception while handling incoming text message.")

    def _check_open(self, ws: WebSocketApp) -> None:
        """
        First signal, that the connection has been opened successfully. Then
        call *self.on_open* and set *self._reader_status* to CHECKING_TOKEN if opening succeeded.

        :param ws: WebSocketApp
        """
        logger.debug("Connection to %s open.", self._url)

        try:
            with self._reader_guard:
                self._reader_status = ReaderStatus.RUNNING_PRELIMINARY

            self.on_open(ws)
        except Exception as err:
            self._set_error(err)
            self._close(status=STATUS_UNEXPECTED_CONDITION,
                        reason="Exception while handling open message.")

    def _check_close(self, ws: WebSocketApp, code: int, reason: str):
        """
        Call *self.on_close*. When *code* and *reason* are None, the stop has been issued locally and not by the
        remote side.

        :param ws: WebSocketApp
        :param code: Code of stop message
        :param reason: Reason str of stop message
        """
        try:
            with self._reader_guard:
                if code or reason:
                    if code in self._remote_exit_codes or self._reader_status in [ReaderStatus.DONE,
                                                                                  ReaderStatus.FAILED]:
                        logger.debug("Received close from remote: %s %s. Closing...", code, reason)
                    else:
                        logger.debug("Received close from remote: %s %s. Restarting...", code, reason)
                        self._reader_status = ReaderStatus.RESTARTING
                else:
                    logger.debug("Received local close. Exiting...")

            self.on_close(ws, code, reason)
        except Exception as err:
            self._set_error(err)

    def _check_error(self, ws: WebSocketApp, error: Exception) -> None:
        """
        Just log the error and propagate it to *self.on_error*.

        :param ws: WebSocketApp
        :param error: Exception
        """
        # logger.error("Received error: %s", str(error))
        # logger.exception(error)

        try:
            # A '[Errno 9] Bad file descriptor' is expected when closing the websocket
            # internally.
            with self._reader_guard:
                if self._reader_status == ReaderStatus.DONE and isinstance(error, OSError) and error.errno == 9:
                    return

            self._set_error(error)
            self.on_error(ws, error)
        except Exception as err:
            self._set_error(err)

    def _backoff(self, reconnect_delay: int) -> int:
        """
        Sleeps for *reconnect_delay* seconds, then returns the delay in seconds for the next try.

        :param reconnect_delay: Delay in seconds to wait.
        :return: Next value for the delay.
        """
        with self._backoff_condition:
            if reconnect_delay:
                self._backoff_condition.wait(timeout=reconnect_delay)

        return (reconnect_delay + 1) if reconnect_delay < 10 \
            else (reconnect_delay + 10) if reconnect_delay < 60 \
            else random.randint(60, 600)

    def _close(self, status: str = STATUS_NORMAL, reason: str = None):
        """
        Close the websocket.

        :param status: Status code for close message.
        :param reason: The close message.
        """
        with self._ws_guard:
            if self._ws:
                self._ws.close(status=status,
                               reason=reason if reason else f"{self._api_handler.user_agent} closing")

    ###############################################################################################################
    # Public API Reader thread
    ###############################################################################################################

    @abstractmethod
    def on_open(self, ws: WebSocketApp):
        pass

    @abstractmethod
    def on_close(self, ws: WebSocketApp, code: int = None, reason: str = None):
        pass

    @abstractmethod
    def on_message(self, ws: WebSocketApp, message: str):
        pass

    @abstractmethod
    def on_error(self, ws: WebSocketApp, error: Exception):
        pass

    ###############################################################################################################
    # Public API Main Writer thread
    ###############################################################################################################

    def start(self) -> None:
        """
        Create the WebSocketApp

        :raise WebSocketException: When the creation of the WebSocketApp fails.
        """

        try:
            with self._ws_guard:
                self._ws = WebSocketApp(self._url,
                                        header={
                                            "Sec-WebSocket-Protocol":
                                                f"{self._protocol}, token-{self._api_handler.token}"
                                        },
                                        on_open=lambda ws: self._check_open(ws),
                                        on_close=lambda ws, code, reason: self._check_close(ws, code, reason),
                                        on_message=lambda ws, msg: self._check_message(ws, msg),
                                        on_error=lambda ws, _err: self._check_error(ws, _err),
                                        on_ping=lambda ws, data: ws.send(data, opcode=ABNF.OPCODE_PONG))

        except Exception as err:
            raise WebSocketException("Cannot create WebSocketApp.") from err

    def _stop(self, wait_for_shutdown: bool = True) -> None:
        """
        Intentionally closes this websocket. When called by the same thread as *self.run_forever()*
        (i.e. by signal handler), *wait_for_shutdown* needs to be set to False or the method will deadlock.

        :param wait_for_shutdown: Wait until *self.run_forever()* has returned. Default is True.
        """
        with self._ws_guard:
            if not self._ws:
                return

        with self._reader_guard:
            self._reader_status = ReaderStatus.DONE
            self._close()
            with self._backoff_condition:
                self._backoff_condition.notify()
            if wait_for_shutdown:
                self._reader_guard.wait()

    def stop(self) -> None:
        """
        Intentionally closes this websocket and waits for *self.run_forever()* to return. Call this from another thread
        than *self.run_forever()*.
        """
        self._stop(wait_for_shutdown=True)

    def signal_stop(self) -> None:
        """
        Intentionally closes this websocket without waiting. This is meant to be used in signal handlers.
        """
        self._stop(wait_for_shutdown=False)

    def restart(self) -> None:
        """
        Closes the websocket and starts a new one. This needs to be called from another thread than
        *self.run_forever()* or it will deadlock.

        :raise RuntimeError: When no *self._ws* exists.
        """
        with self._ws_guard:
            if not self._ws:
                raise RuntimeError('There is no websocket to restart.')

        with self._reader_guard:
            self._reader_status = ReaderStatus.RESTARTING
            self._close()
            with self._backoff_condition:
                self._backoff_condition.notify()
            self._reader_guard.wait()

    def run_forever(self):
        """
        Runs and receives incoming messages.
        """
        with self._reader_guard:
            self._reader_status = ReaderStatus.STARTING
            self._reconnect_delay = 0

            while self._reader_status not in [ReaderStatus.DONE, ReaderStatus.FAILED]:

                try:
                    self._reader_guard.notify_all()
                    self._reader_guard.release()
                    self._ws.run_forever(http_proxy_host=self._proxy_hostname,
                                         http_proxy_port=self._proxy_port,
                                         http_proxy_auth=self._proxy_auth,
                                         sslopt={
                                             "cert_reqs": ssl.CERT_NONE
                                         } if AbstractTokenApiHandler.accept_all_certs else None)

                except Exception as error:
                    self._check_error(self._ws, error)
                finally:
                    self._reader_guard.acquire()

                self._reconnect_delay = self._backoff(self._reconnect_delay)

            self.on_close(self._ws)

            with self._ws_guard:
                self._ws = None

            self._reader_guard.notify_all()

    def is_active(self) -> bool:
        """
        Checks whether the websocket is still active.
        :return: self._reader_status not in [ReaderStatus.NONE, ReaderStatus.DONE, ReaderStatus.FAILED]
        """
        with self._reader_guard:
            return self._reader_status not in [ReaderStatus.NONE, ReaderStatus.DONE, ReaderStatus.FAILED]

    def send(self, message: str) -> None:
        """
        Send message across the websocket.

        :param message: Message as string
        :raise WebSocketException: When *self._auto_reconnect* is False: If a message cannot be sent and all retries
                                   have been exhausted.
        :raise WebSocketConnectionClosedException: When the websocket is not available at all.
        """
        retries = 0
        retry_delay = 0

        while True:
            retry_delay = self._backoff(retry_delay)

            try:
                with self._reader_guard:
                    if self._reader_status in [ReaderStatus.NONE]:
                        raise WebSocketConnectionClosedException('Websocket not started.')
                    elif self._reader_status in [ReaderStatus.DONE, ReaderStatus.FAILED]:
                        raise WebSocketConnectionClosedException('Websocket has exited.')
                    elif self._reader_status not in [ReaderStatus.RUNNING, ReaderStatus.RUNNING_PRELIMINARY]:
                        raise WebSocketException('Websocket not ready.')

                with self._ws_guard:
                    if not self._ws:
                        raise WebSocketConnectionClosedException('Websocket is gone.')

                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Sending message: %s", message)
                    else:
                        logger.info("Sending message of length %d", len(message))

                    self._ws.send(message)

                return

            except WebSocketConnectionClosedException:
                raise

            except Exception as err:
                if retries >= self.MAX_RETRIES:
                    if self._auto_reconnect:
                        retries = 0
                        logger.warning('Restarting because of error: %s', str(err))
                        self.restart()
                    else:
                        raise WebSocketException("Could not send and all retries have been exhausted.")
                else:
                    logger.warning('Retrying to send message because of error: %s', str(err))
                    retries += 1



#!/usr/bin/env python3
import json

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroKi(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO Ki REST API.
    See https://core.arago.co/help/specs/?url=definitions/ki.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='ki',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations against the ki API
    ###############################################################################################################

    def check(self, data: dict) -> dict:
        """
        Ask if KI is valid

        HIRO REST query API: `POST self._endpoint + '/check`

        :param data: KI validation request data.
               See https://core.arago.co/help/specs/?url=definitions/ki.yaml#/[Validatation]_Validate/post_check
        :return: The result payload
        """
        url = self.endpoint + '/check'
        return self.post(url, data)

    ###############################################################################################################
    # Internal methods
    ###############################################################################################################

    def _get_error_message(self, json_result: dict) -> str:
        """
        Intercept special error messages. These have a key 'status' in their dict.

        :param json_result: The incoming JSON containing error information.
        :return: The error message.
        """
        if 'status' in json_result:
            return json.dumps(json_result)

        return super()._get_error_message(json_result)
